#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Implement DataFile like classes to represent Instrument Manufacturer's File Formats'."""

# Standard Library imports
from datetime import datetime
import re
import struct
from ast import literal_eval

import numpy as np

from ...compat import str2bytes, bytes2str
from ...core.exceptions import StonerAssertionError, assertion, StonerLoadError
from ...core.base import string_to_type
from ...tools.file import FileManager, SizedFileManager

from ..decorators import register_loader, register_saver


@register_loader(patterns=(".340", 16), mime_types=("text/plain", 32), name="LSTemperatureFile", what="Data")
def load_340(new_data, filename=None, *args, **kargs):
    """Load data for 340 files."""
    new_data.filename = filename
    with FileManager(new_data.filename, "rb") as data:
        keys = []
        vals = []
        for line in data:
            line = bytes2str(line)
            if line.strip() == "":
                break
            parts = [p.strip() for p in line.split(":")]
            if len(parts) != 2:
                raise StonerLoadError(f"Header doesn't contain two parts at {line.strip()}")
            keys.append(parts[0])
            vals.append(parts[1])
        else:
            raise StonerLoadError("Overan the end of the file")
        if keys != [
            "Sensor Model",
            "Serial Number",
            "Data Format",
            "SetPoint Limit",
            "Temperature coefficient",
            "Number of Breakpoints",
        ]:
            raise StonerLoadError("Header did not contain recognised keys.")
        for k, v in zip(keys, vals):
            v = v.split()[0]
            new_data.metadata[k] = string_to_type(v)
        headers = bytes2str(next(data)).strip().split()
        column_headers = headers[1:]
        dat = np.genfromtxt(data)
        new_data.data = dat[:, 1:]
    new_data.column_headers = column_headers
    return new_data


@register_saver(patterns=".340", name="LSTemperatureFile", what="Data")
def save_340(save_data, filename=None, **kargs):
    """Override the save method to allow CSVFiles to be written out to disc (as a mininmalist output).

    Args:
        filename (string):
            Filename to save as (using the same rules as for the load routines)

    Keyword Arguments:
        deliminator (string):
            Record deliniminator (defaults to a comma)

    Returns:
        A copy of itsave_data.
    """
    if filename is None:
        filename = save_data.filename
    if filename is None or (isinstance(filename, bool) and not filename):  # now go and ask for one
        filename = save_data.__file_dialog("w")
    if save_data.shape[1] == 2:  # 2 columns, let's hope they're the right way round!
        cols = [0, 1]
    elif (
        save_data.setas.has_xcol and save_data.setas.has_ycol
    ):  # Use ycol, x col but assume x is real temperature and y is resistance
        cols = [save_data.setas.ycol[0], save_data.setas.xcol]
    else:
        cols = range(save_data.shape[1])
    with FileManager(filename, "w", errors="ignore", encoding="utf-8", newline="\r\n") as f:
        for k, v in (
            ("Sensor Model", "CX-1070-SD"),
            ("Serial Number", "Unknown"),
            ("Data Format", 4),
            ("SetPoint Limit", 300.0),
            ("Temperature coefficient", 1),
            ("Number of Breakpoints", len(save_data)),
        ):
            if k in ["Sensor Model", "Serial Number", "Data Format", "SetPoint Limit"]:
                kstr = f"{k+':':16s}"
            else:
                kstr = f"{k}:   "
            v = save_data.get(k, v)
            if k == "Data Format":
                units = ["()", "()", "()", "()", "(Log Ohms/Kelvin)", "(Log Ohms/Log Kelvin)"]
                vstr = f"{v}      {units[int(v)]}"
            elif k == "SetPointLimit":
                vstr = f"{v}      (Kelvin)"
            elif k == "Temperature coefficient":
                vstr = f"{v} {['(positive)', '(negative)'][v]}"
            elif k == "Number of Breakpoints":
                vstr = str(len(save_data))
            else:
                vstr = str(v)
            f.write(f"{kstr}{vstr}\n")
        f.write("\n")
        f.write("No.   ")
        for i in cols:
            f.write(f"{save_data.column_headers[i]:11s}")
        f.write("\n\n")
        for i in range(
            len(save_data.data)
        ):  # This is a slow way to write the data, but there should only ever be 200 lines
            line = "\t".join([f"{n:<10.8f}" for n in save_data.data[i, cols]])
            f.write(f"{i}\t")
            f.write(f"{line}\n")
    save_data.filename = filename
    return save_data


@register_loader(
    patterns=(".dat", 16),
    mime_types=[("application/x-wine-extension-ini", 15), ("text/plain", 16)],
    name="QDFile",
    what="Data",
)
def load_qdfile(new_data, filename=None, *args, **kargs):
    """QD system file loader routine.

    Args:
        filename (string or bool):
            File to load. If None then the existing filename is used, if False, then a file dialog will be used.

    Returns:
        A copy of the itnew_data after loading the data.
    """
    setas = {}
    i = 0
    new_data.filename = filename
    with FileManager(new_data.filename, "r", encoding="utf-8", errors="ignore") as f:  # Read filename linewise
        for i, line in enumerate(f):
            line = line.strip()
            if i == 0 and line != "[Header]":
                raise StonerLoadError("Not a Quantum Design File !")
            if line == "[Header]" or line.startswith(";") or line == "":
                continue
            if "[Data]" in line:
                break
            if "," not in line:
                raise StonerLoadError("No data in file!")
            parts = [x.strip() for x in line.split(",")]
            if parts[1].split(":")[0] == "SEQUENCE FILE":
                key = parts[1].split(":")[0].title()
                value = parts[1].split(":")[1]
            elif parts[0] == "INFO":
                if parts[1] == "APPNAME":
                    parts[1], parts[2] = parts[2], parts[1]
                if len(parts) > 2:
                    key = f"{parts[0]}.{parts[2]}"
                else:
                    raise StonerLoadError("No data in file!")
                key = key.title()
                value = parts[1]
            elif parts[0] in ["BYAPP", "FILEOPENTIME"]:
                key = parts[0].title()
                value = " ".join(parts[1:])
            elif parts[0] == "FIELDGROUP":
                key = f"{parts[0]}.{parts[1]}".title()
                value = f'[{",".join(parts[2:])}]'
            elif parts[0] == "STARTUPAXIS":
                axis = parts[1][0].lower()
                setas[axis] = setas.get(axis, []) + [int(parts[2])]
                key = f"Startupaxis-{parts[1].strip()}"
                value = parts[2].strip()
            else:
                key = parts[0] + "," + parts[1]
                key = key.title()
                value = " ".join(parts[2:])
            new_data.metadata[key] = string_to_type(value)
        else:
            raise StonerLoadError("No data in file!")
        if "Byapp" not in new_data:
            raise StonerLoadError("Not a Quantum Design File !")

        column_headers = f.readline().strip().split(",")
        data = np.genfromtxt([str2bytes(l) for l in f], dtype="float", delimiter=",", invalid_raise=False)
        if data.shape[0] == 0:
            raise StonerLoadError("No data in file!")
        if data.shape[1] < len(column_headers):  # Trap for buggy QD software not giving ewnough columns of data
            data = np.append(data, np.ones((data.shape[0], len(column_headers) - data.shape[1])) * np.NaN, axis=1)
        elif data.shape[1] > len(column_headers):  # too much data
            data = data[:, : len(column_headers) - data.shape[1]]
        new_data.data = data
    new_data.column_headers = column_headers
    s = new_data.setas
    for k in setas:
        for ix in setas[k]:
            s[ix - 1] = k
    new_data.setas = s
    return new_data


def _to_Q(new_data, wavelength=1.540593):
    """Add an additional function to covert an angualr scale to momentum transfer.

    Returns:
        a copy of itnew_data.
    """
    new_data.add_column(
        (4 * np.pi / wavelength) * np.sin(np.pi * new_data.column(0) / 360), header="Momentum Transfer, Q ($\\AA$)"
    )


@register_loader(
    patterns=(".ras", 16),
    mime_types=[("application/x-wine-extension-ini", 16), ("text/plain", 16)],
    name="RigakuFile",
    what="Data",
)
def load_rigaku(new_data, filename=None, *args, **kargs):
    """Read a Rigaku ras file including handling the metadata nicely.

    Args:
        filename (string or bool):
            File to load. If None then the existing filename is used, if False, then a file dialog will be used.

    Returns:
        A copy of the itnew_data after loading the data.
    """
    sh = re.compile(r"^\*([^\s]+)\s+(.*)$")  # Regexp to grab the keys
    ka = re.compile(r"(.*)\-(\d+)$")
    header = {}
    i = 0
    new_data.filename = filename
    with SizedFileManager(new_data.filename, "rb") as (f, end):
        for i, line in enumerate(f):
            line = bytes2str(line).strip()
            if i == 0 and not line.startswith("*RAS_"):
                raise StonerLoadError("Not a Rigaku file!")
            if line == "*RAS_HEADER_START":
                break
        for line in f:
            line = bytes2str(line).strip()
            m = sh.match(line)
            if m:
                key = m.groups()[0].lower().replace("_", ".")
                try:
                    value = m.groups()[1].decode("utf-8", "ignore")
                except AttributeError:
                    value = m.groups()[1]
                header[key] = value
            if "*RAS_INT_START" in line:
                break
        keys = list(header.keys())
        keys.sort()
        for key in keys:
            m = ka.match(key)
            value = header[key].strip()
            try:
                newvalue = literal_eval(value.strip('"'))
            except (TypeError, ValueError, SyntaxError):
                newvalue = literal_eval(value)
            if newvalue == "-":
                newvalue = np.nan  # trap for missing float value
            if m:
                key = m.groups()[0]
                idx = int(m.groups()[1])
                if key in new_data.metadata and not (isinstance(new_data[key], (np.ndarray, list))):
                    if isinstance(new_data[key], str):
                        new_data[key] = list([new_data[key]])
                        if idx > 1:
                            new_data[key].extend([""] * idx - 1)
                    else:
                        new_data[key] = np.array(new_data[key])
                        if idx > 1:
                            new_data[key] = np.append(new_data[key], np.ones(idx - 1) * np.nan)
                if key not in new_data.metadata:
                    if isinstance(newvalue, str):
                        listval = [""] * (idx + 1)
                        listval[idx] = newvalue
                        new_data[key] = listval
                    else:
                        arrayval = np.ones(idx + 1) * np.nan
                        arrayval = arrayval.astype(type(newvalue))
                        arrayval[idx] = newvalue
                        new_data[key] = arrayval
                else:
                    if isinstance(new_data[key][0], str) and isinstance(new_data[key], list):
                        if len(new_data[key]) < idx + 1:
                            new_data[key].extend([""] * (idx + 1 - len(new_data[key])))
                        new_data[key][idx] = newvalue
                    else:
                        if idx + 1 > new_data[key].size:
                            new_data[key] = np.append(
                                new_data[key],
                                (np.ones(idx + 1 - new_data[key].size) * np.nan).astype(new_data[key].dtype),
                            )
                        try:
                            new_data[key][idx] = newvalue
                        except ValueError:
                            pass
            else:
                new_data.metadata[key] = newvalue

        pos = f.tell()
        max_rows = 0
        for max_rows, line in enumerate(f):
            line = bytes2str(line).strip()
            if "RAS_INT_END" in line:
                break
        f.seek(pos)
        if max_rows > 0:
            new_data.data = np.genfromtxt(
                f, dtype="float", delimiter=" ", invalid_raise=False, comments="*", max_rows=max_rows
            )
            column_headers = ["Column" + str(i) for i in range(new_data.data.shape[1])]
            column_headers[0:2] = [new_data.metadata["meas.scan.unit.x"], new_data.metadata["meas.scan.unit.y"]]
            for key in new_data.metadata:
                if isinstance(new_data[key], list):
                    new_data[key] = np.array(new_data[key])
            new_data.setas = "xy"
            new_data.column_headers = column_headers
        pos = f.tell()
    if pos < end:  # Trap for Rigaku files with multiple scans in them.
        new_data["_endpos"] = pos
        if hasattr(filename, "seekable") and filename.seekable():
            filename.seek(pos)
    if kargs.pop("add_Q", False):
        _to_Q(new_data)
    return new_data


def _read_spc_xdata(new_data, f):
    """Read the xdata from the spc file."""
    new_data._pts = new_data._header["fnpts"]
    if new_data._header["ftflgs"] & 128:  # We need to read some X Data
        if 4 * new_data._pts > new_data._filesize - f.tell():
            raise StonerLoadError("Trying to read too much data!")
        xvals = f.read(4 * new_data._pts)  # I think storing X vals directly implies that each one is 4 bytes....
        xdata = np.array(struct.unpack(str2bytes(str(new_data._pts) + "f"), xvals))
    else:  # Generate the X Data ourselves
        first = new_data._header["ffirst"]
        last = new_data._header["flast"]
        if new_data._pts > 1e6:  # Something not right here !
            raise StonerLoadError("More than 1 million points requested. Bugging out now!")
        xdata = np.linspace(first, last, new_data._pts)
    return xdata


def _read_spc_ydata(new_data, f, data, column_headers):
    """Read the y data and column headers from spc file."""
    n = new_data._header["fnsub"]
    subhdr_keys = (
        "subflgs",
        "subexp",
        "subindx",
        "subtime",
        "subnext",
        "subnois",
        "subnpts",
        "subscan",
        "subwlevel",
        "subresv",
    )
    if new_data._header["ftflgs"] & 1:
        y_width = 2
        y_fmt = "h"
        divisor = 2**16
    else:
        y_width = 4
        y_fmt = "i"
        divisor = 2**32
    if n * (y_width * new_data._pts + 32) > new_data._filesize - f.tell():
        raise StonerLoadError("No good, going to read too much data!")
    for j in range(n):  # We have n sub-scans
        # Read the subheader and import into the main metadata dictionary as scan#:<subheader item>
        subhdr = struct.unpack(b"BBHfffIIf4s", f.read(32))
        subheader = dict(zip(["scan" + str(j) + ":" + x for x in subhdr_keys], subhdr))

        # Now read the y-data
        exponent = subheader["scan" + str(j) + ":subexp"]
        if int(exponent) & -128:  # Data is unscaled direct floats
            ydata = np.array(struct.unpack(str2bytes(str(new_data._pts) + "f"), f.read(new_data._pts * y_width)))
        else:  # Data is scaled by exponent
            yvals = struct.unpack(str2bytes(str(new_data._pts) + y_fmt), f.read(new_data._pts * y_width))
            ydata = np.array(yvals, dtype="float64") * (2**exponent) / divisor
        data[:, j + 1] = ydata
        new_data._header = dict(new_data._header, **subheader)
        column_headers.append("Scan" + str(j) + ":" + new_data._yvars[new_data._header["fytype"]])

    return data


def _read_spc_loginfo(new_data, f):
    """Read the log info section of the spc file."""
    logstc = struct.unpack(b"IIIII44s", f.read(64))
    logstc_keys = ("logsizd", "logsizm", "logtxto", "logbins", "logdsks", "logrsvr")
    logheader = dict(zip(logstc_keys, logstc))
    new_data._header = dict(new_data._header, **logheader)

    # Can't handle either binary log information or ion disk log information (wtf is this anyway !)
    if new_data._header["logbins"] + new_data._header["logdsks"] > new_data._filesize - f.tell():
        raise StonerLoadError("Too much logfile data to read")
    f.read(new_data._header["logbins"] + new_data._header["logdsks"])

    # The renishaw seems to put a 16 character timestamp next - it's not in the spec but never mind that.
    new_data._header["Date-Time"] = f.read(16)
    # Now read the rest of the file as log text
    logtext = f.read()
    # We expect things to be single lines terminated with a CR-LF of the format key=value
    for line in re.split(b"[\r\n]+", logtext):
        if b"=" in line:
            parts = line.split(b"=")
            key = parts[0].decode()
            value = parts[1].decode()
            new_data._header[key] = value


@register_loader(patterns=(".spc", 16), mime_types=("application/octet-stream", 16), name="SPCFile", what="Data")
def load_spc(new_data, filename=None, *args, **kargs):
    """Read a .scf file produced by the Renishaw Raman system (among others).

    Args:
        filename (string or bool):
            File to load. If None then the existing filename is used, if False, then a file dialog will be used.

    Returns:
        A copy of the itnew_data after loading the data.

    Todo:
        Implement the second form of the file that stores multiple x-y curves in the one file.

    Notes:
        Metadata keys are pretty much as specified in the spc.h file that defines the filerformat.
    """
    if filename is None or not filename:
        new_data.get_filename("r")
    else:
        new_data.filename = filename
    # Open the file and read the main file header and unpack into a dict
    with SizedFileManager(filename, "rb") as (f, length):
        new_data._filesize = length
        spchdr = struct.unpack(b"BBBciddiBBBBi9s9sH8f30s130siiBBHf48sfifB187s", f.read(512))
        keys = (
            "ftflgs",
            "fversn",
            "fexper",
            "fexp",
            "fnpts",
            "ffirst",
            "flast",
            "fnsub",
            "fxtype",
            "fytype",
            "fztype",
            "fpost",
            "fres",
            "fsource",
            "fpeakpt",
            "fspare1",
            "fspare2",
            "fspare3",
            "fspare4",
            "fspare5",
            "fspare6",
            "fspare7",
            "fspare8",
            "fcm",
            "nt",
            "fcatx",
            "flogoff",
            "fmods",
            "fprocs",
            "flevel",
            "fsampin",
            "ffactor",
            "fmethod",
            "fzinc",
            "fwplanes",
            "fwinc",
            "fwtype",
            "fwtype",
            "fresv",
        )
        new_data._xvars = [
            "Arbitrary",
            "Wavenumber (cm-1)",
            "Micrometers (um)",
            "Nanometers (nm)",
            "Seconds",
            "Minutes",
            "Hertz (Hz)",
            "Kilohertz (KHz)",
            "Megahertz (MHz)",
            "Mass (M/z)",
            "Parts per million (PPM)",
            "Days",
            "Years",
            "Raman Shift (cm-1)",
            "Raman Shift (cm-1)",
            "eV",
            "XYZ text labels in fcatxt (old 0x4D version only)",
            "Diode Number",
            "Channel",
            "Degrees",
            "Temperature (F)",
            "Temperature (C)",
            "Temperature (K)",
            "Data Points",
            "Milliseconds (mSec)",
            "Microseconds (uSec)",
            "Nanoseconds (nSec)",
            "Gigahertz (GHz)",
            "Centimeters (cm)",
            "Meters (m)",
            "Millimeters (mm)",
            "Hours",
            "Hours",
        ]
        new_data._yvars = [
            "Arbitrary Intensity",
            "Interferogram",
            "Absorbance",
            "Kubelka-Monk",
            "Counts",
            "Volts",
            "Degrees",
            "Milliamps",
            "Millimeters",
            "Millivolts",
            "Log(1/R)",
            "Percent",
            "Percent",
            "Intensity",
            "Relative Intensity",
            "Energy",
            "Decibel",
            "Temperature (F)",
            "Temperature (C)",
            "Temperature (K)",
            "Index of Refraction [N]",
            "Extinction Coeff. [K]",
            "Real",
            "Imaginary",
            "Complex",
            "Complex",
            "Transmission (ALL HIGHER MUST HAVE VALLEYS!)",
            "Reflectance",
            "Arbitrary or Single Beam with Valley Peaks",
            "Emission",
            "Emission",
        ]

        new_data._header = dict(zip(keys, spchdr))
        n = new_data._header["fnsub"]

        if new_data._header["ftflgs"] & 64 == 64 or not (
            75 <= new_data._header["fversn"] <= 77
        ):  # This is the multiple XY curves in file flag.
            raise StonerLoadError(
                f"Filetype not implemented yet ! {new_data._header['ftflgs']=}, {new_data._header['fversn']=}"
            )
        # Read the xdata and add it to the file.
        xdata = _read_spc_xdata(new_data, f)
        data = np.zeros((new_data._pts, (n + 1)))  # initialise the data soace
        data[:, 0] = xdata  # Put in the X-Data
        column_headers = [new_data._xvars[new_data._header["fxtype"]]]  # And label the X column correctly

        # Now we're going to read the Y-data
        data = _read_spc_ydata(new_data, f, data, column_headers)
        if new_data._header["flogoff"] != 0:  # Ok, we've got a log, so read the log header and merge into metadata
            _read_spc_loginfo(new_data, f)
        # Ok now build the Stoner.DataFile instance to return
        new_data.data = data
        # The next bit generates the metadata. We don't just copy the metadata because we need to figure out
        # the typehints first - hence the loop
        # here to call DataFile.__setitem()
        for x in new_data._header:
            new_data[x] = new_data._header[x]
        new_data.column_headers = column_headers
        if len(new_data.column_headers) == 2:
            new_data.setas = "xy"
        return new_data


@register_loader(patterns=[(".fld", 16), (".dat", 32)], mime_types=("text/plain", 16), name="VSMFile", what="Data")
def load_vsm(new_data, filename=None, *args, header_line=3, data_line=3, header_delim=",", **kargs):
    """VSM file loader routine.

    Args:
        filename (string or bool):
            File to load. If None then the existing filename is used, if False, then a file dialog will be used.

    Keyword Arguments:
        header_line (int):
            The line in the file that contains the column headers. If None, then column headers are automatically
            generated.
        data_line (int):
            The line on which the data starts
        header_delim (strong):
            The delimiter used for separating header values

    Returns:
        A copy of the itnew_data after loading the data.
    """
    new_data.filename = filename
    try:
        with FileManager(filename, errors="ignore", encoding="utf-8") as f:
            for i, line in enumerate(f):
                if i == 0:
                    first = line.strip()
                    new_data["Timestamp"] = first
                    check = datetime.strptime(first, "%a %b %d %H:%M:%S %Y")
                    if check is None:
                        raise StonerLoadError("Not a VSM file ?")
                elif i == 1:
                    assertion(line.strip() == "")
                elif i == 2:
                    header_string = line.strip()
                elif i == header_line:
                    unit_string = line.strip()
                    column_headers = [
                        f"{h.strip()} ({u.strip()})"
                        for h, u in zip(header_string.split(header_delim), unit_string.split(header_delim))
                    ]
                elif i > 3:
                    break
    except (StonerAssertionError, ValueError, AssertionError, TypeError) as err:
        raise StonerLoadError(f"Not a VSM File {err}") from err
    new_data.data = np.genfromtxt(
        new_data.filename,
        dtype="float",
        usemask=True,
        skip_header=data_line - 1,
        missing_values=["6:0", "---"],
        invalid_raise=False,
    )

    new_data.data = np.ma.mask_rows(new_data.data)
    cols = new_data.data.shape[1]
    new_data.data = np.reshape(new_data.data.compressed(), (-1, cols))
    new_data.column_headers = column_headers
    new_data.setas(x="H_vsm (T)", y="m (emu)")  # pylint: disable=not-callable
    return new_data


@register_loader(
    patterns=".dql",
    mime_types=[("application/x-wine-extension-ini", 16), ("text/plain", 16)],
    name="XRDFile",
    what="Data",
)
def load_xrd(new_data, filename=None, *args, **kargs):
    """Read an XRD DataFile as produced by the Brucker diffractometer.

    Args:
        filename (string or bool):
            File to load. If None then the existing filename is used, if False, then a file dialog will be used.

    Returns:
        A copy of the itnew_data after loading the data.

    Notes:
        Format is ini file like but not enough to do standard inifile processing - in particular
        one can have multiple sections with the same name (!)
    """
    if filename is None or not filename:
        new_data.get_filename("r")
    else:
        new_data.filename = filename
    sh = re.compile(r"\[(.+)\]")  # Regexp to grab section name
    with FileManager(new_data.filename, errors="ignore", encoding="utf-8") as f:  # Read filename linewise
        if f.readline().strip() != ";RAW4.00":  # Check we have the correct fileformat
            raise StonerLoadError("File Format Not Recognized !")
        drive = 0
        for line in f:  # for each line
            m = sh.search(line)
            if m:  # This is a new section
                section = m.group(1)
                if section == "Drive":  # If this is a Drive section we need to know which Drive Section it is
                    section = section + str(drive)
                    drive = drive + 1
                elif section == "Data":  # Data section contains the business but has a redundant first line
                    f.readline()
                for line in f:  # Now start reading lines in this section...
                    if line.strip() == "":
                        # A blank line marks the end of the section, so go back to the outer loop which will
                        # handle a new section
                        break
                    if section == "Data":  # In the Data section read lines of data value,vale
                        parts = line.split(",")
                        angle = parts[0].strip()
                        counts = parts[1].strip()
                        dataline = np.array([float(angle), float(counts)])
                        new_data.data = np.append(new_data.data, dataline)
                    else:  # Other sections contain metadata
                        parts = line.split("=")
                        key = parts[0].strip()
                        data = parts[1].strip()
                        # Keynames in main metadata are section:key - use theDataFile magic to do type
                        # determination
                        new_data[section + ":" + key] = string_to_type(data)
        column_headers = ["Angle", "Counts"]  # Assume the columns were Angles and Counts

    new_data.data = np.reshape(new_data.data, (-1, 2))
    new_data.setas = "xy"
    new_data._public_attrs = {"four_bounce": bool}
    new_data.four_bounce = new_data["HardwareConfiguration:Monochromator"] == 1
    new_data.column_headers = column_headers
    if kargs.pop("Q", False):
        _to_Q(new_data)
    return new_data



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Implement DataFile like classes for Various experimental rigs."""

import re
import csv

import numpy as np

from ...compat import bytes2str
from ...core.base import string_to_type
from ...core.exceptions import StonerLoadError
from ..data.generic import load_csvfile
from ...tools.file import FileManager

from ..decorators import register_loader


@register_loader(
    patterns=[(".dat", 64), (".iv", 64), (".rvt", 64)], mime_types=("text/plain", 64), name="BigBlueFile", what="Data"
)
def load_bigblue(new_data, filename, *args, **kargs):
    """Just call the parent class but with the right parameters set.

    Args:
        filename (string or bool): File to load. If None then the existing filename is used,
            if False, then a file dialog will be used.

    Returns:
        A copy of the itnew_data after loading the data.
    """
    new_data.filename = filename

    new_data = load_csvfile(new_data, filename, *args, header_line=3, data_line=7, data_delim=" ", header_delim=",")
    if np.all(np.isnan(new_data.data)):
        raise StonerLoadError("All data was NaN in Big Blue format")
    return new_data


@register_loader(patterns=(".dat", 32), mime_types=("text/plain", 32), name="BirgeIVFile", what="Data")
def load_birge(new_data, filename, *args, **kargs):
    """File loader for PinkLib.

    Args:
        filename (string or bool): File to load. If None then the existing filename is used,
            if False, then a file dialog will be used.

    Returns:
        A copy of the itnew_data after loading the data.
    """
    new_data.filename = filename
    ix = 0
    with FileManager(new_data.filename, "r", errors="ignore", encoding="utf-8") as f:  # Read filename linewise
        if not re.compile(r"\d{1,2}/\d{1,2}/\d{4}").match(f.readline()):
            raise StonerLoadError("Not a BirgeIVFile as no date on first line")
        data = f.readlines()
        expected = ["Vo(-))", "Vo(+))", "Ic(+)", "Ic(-)"]
        for length, m in zip(data[-4:], expected):
            if not length.startswith(m):
                raise StonerLoadError("Not a BirgeIVFile as wrong footer line")
            key = length[: len(m)]
            val = length[len(m) :]
            if "STDEV" in val:
                ix2 = val.index("STDEV")
                key2 = val[ix2 : ix2 + 4 + len(key)]
                val2 = val[ix2 + 4 + len(key) :]
                new_data.metadata[key2] = string_to_type(val2.strip())
                val = val[:ix2]
            new_data.metadata[key] = string_to_type(val.strip())
        for ix, line in enumerate(data):  # Scan the ough lines to get metadata
            if ":" in line:
                parts = line.split(":")
                new_data.metadata[parts[0].strip()] = string_to_type(parts[1].strip())
            elif "," in line:
                for part in line.split(","):
                    parts = part.split(" ")
                    new_data.metadata[parts[0].strip()] = string_to_type(parts[1].strip())
            elif line.startswith("H "):
                new_data.metadata["H"] = string_to_type(line.split(" ")[1].strip())
            else:
                headers = [x.strip() for x in line.split(" ")]
                break
        else:
            raise StonerLoadError("Oops ran off the end of the file!")
    new_data.data = np.genfromtxt(filename, skip_header=ix + 2, skip_footer=4)
    new_data.column_headers = headers

    new_data.setas = "xy"
    return new_data


@register_loader(patterns=[(".dat", 16), (".txt", 16)], mime_types=("text/plain", 16), name="MokeFile", what="Data")
def load_old_moke(new_data, filename, *args, **kargs):
    """Leeds  MOKE file loader routine.

    Args:
        filename (string or bool): File to load. If None then the existing filename is used,
            if False, then a file dialog will be used.

    Returns:
        A copy of the itnew_data after loading the data.
    """
    new_data.filename = filename
    with FileManager(new_data.filename, mode="rb") as f:
        line = bytes2str(f.readline()).strip()
        if line != "#Leeds CM Physics MOKE":
            raise StonerLoadError("Not a Core.DataFile from the Leeds MOKE")
        while line.startswith("#") or line == "":
            parts = line.split(":")
            if len(parts) > 1:
                key = parts[0][1:]
                data = ":".join(parts[1:]).strip()
                new_data[key] = data
            line = bytes2str(f.readline()).strip()
        column_headers = [x.strip() for x in line.split(",")]
        new_data.data = np.genfromtxt(f, delimiter=",")
    new_data.setas = "xy.de"
    new_data.column_headers = column_headers
    return new_data


@register_loader(patterns=(".dat", 16), mime_types=("text/plain", 16), name="FmokeFile", what="Data")
def load_fmoke(new_data, filename, *args, **kargs):
    """Sheffield Focussed MOKE file loader routine.

    Args:
        filename (string or bool): File to load. If None then the existing filename is used,
            if False, then a file dialog will be used.

    Returns:
        A copy of the itnew_data after loading the data.
    """
    if filename is None or not filename:
        new_data.get_filename("r")
    else:
        new_data.filename = filename
    with FileManager(new_data.filename, mode="rb") as f:
        try:
            value = [float(x.strip()) for x in bytes2str(f.readline()).split("\t")]
        except (TypeError, ValueError) as err:
            f.close()
            raise StonerLoadError("Not an FMOKE file?") from err
        label = [x.strip() for x in bytes2str(f.readline()).split("\t")]
        if label[0] != "Header:":
            f.close()
            raise StonerLoadError("Not a Focussed MOKE file !")
        del label[0]
        for k, v in zip(label, value):
            new_data.metadata[k] = v  # Create metadata from first 2 lines
        column_headers = [x.strip() for x in bytes2str(f.readline()).split("\t")]
        new_data.data = np.genfromtxt(f, dtype="float", delimiter="\t", invalid_raise=False)
        new_data.column_headers = column_headers
    return new_data


def _extend_columns(new_data, i):
    """Ensure the column headers are at least i long."""
    if len(new_data.column_headers) < i:
        length = len(new_data.column_headers)
        new_data.data = np.append(
            new_data.data, np.zeros((new_data.shape[0], i - length)), axis=1
        )  # Need to expand the array first
        new_data.column_headers.extend([f"Column {x}" for x in range(length, i)])


def _et_cmd(new_data, parts):
    """Handle axis labellling command."""
    if parts[0] == "x":
        _extend_columns(new_data, 1)
        new_data.column_headers[0] = parts[1]
    elif parts[0] == "y":
        _extend_columns(new_data, 2)
        new_data.column_headers[1] = parts[1]
    elif parts[0] == "g":
        new_data["title"] = parts[1]


def _td_cmd(new_data, parts):
    new_data.setas = parts[0]


def _sa_cmd(new_data, parts):
    """Implement the sa (set-axis?) command."""
    if parts[0] == "l":  # Legend
        col = int(parts[2])
        _extend_columns(new_data, col + 1)
        new_data.column_headers[col] = parts[1]


@register_loader(patterns=("*", 64), mime_types=("text/plain", 64), name="EasyPlotFile", what="Data")
def load_easyplot(new_data, filename, *args, **kargs):
    """Private loader method."""
    new_data.filename = filename

    datastart = -1
    dataend = -1

    i = 0
    with FileManager(new_data.filename, "r", errors="ignore", encoding="utf-8") as data:
        if "******** EasyPlot save file ********" not in data.read(1024):
            raise StonerLoadError("Not an EasyPlot Save file?")
        data.seek(0)
        for i, line in enumerate(data):
            line = line.strip()
            if line == "":
                continue
            if line[0] not in "-0123456789" and dataend < 0 <= datastart:
                dataend = i
            if line.startswith('"') and ":" in line:
                parts = [x.strip() for x in line.strip('"').split(":")]
                new_data[parts[0]] = string_to_type(":".join(parts[1:]))
            elif line.startswith("/"):  # command
                parts = [x.strip('"') for x in next(csv.reader([line], delimiter=" ")) if x != ""]
                cmd = parts[0].strip("/")
                if len(cmd) > 1:
                    cmdname = f"_{cmd}_cmd"
                    if cmdname in globals():
                        cmd = globals()[cmdname]
                        cmd(new_data, parts[1:])
                    else:
                        if len(parts[1:]) > 1:
                            cmd = cmd + "." + parts[1]
                            value = ",".join(parts[2:])
                        elif len(parts[1:]) == 1:
                            value = parts[1]
                        else:
                            value = True
                        new_data[cmd] = value
            elif line[0] in "-0123456789" and datastart < 0:  # start of data
                datastart = i
                if "," in line:
                    delimiter = ","
                else:
                    delimiter = None
    if dataend < 0:
        dataend = i
    new_data.data = np.genfromtxt(
        new_data.filename, skip_header=datastart, skip_footer=i - dataend, delimiter=delimiter
    )
    if new_data.data.shape[1] == 2:
        new_data.setas = "xy"
    return new_data


@register_loader(patterns=(".dat", 64), mime_types=("text/plain", 64), name="PinkLibFile", what="Data")
def load_pinklib(new_data, filename=None, *args, **kargs):
    """File loader for PinkLib.

    Args:
        filename (string or bool): File to load. If None then the existing filename is used,
            if False, then a file dialog will be used.

    Returns:
        A copy of the itnew_data after loading the data.
    """
    if filename is None or not filename:
        new_data.get_filename("r")
    else:
        new_data.filename = filename
    with FileManager(new_data.filename, "r", errors="ignore", encoding="utf-8") as f:  # Read filename linewise
        if "PINKlibrary" not in f.readline():
            raise StonerLoadError("Not a PINK file")
        f = f.readlines()
        happened_before = False
        for i, line in enumerate(f):
            if line[0] != "#" and not happened_before:
                header_line = i - 2  # -2 because there's a commented out data line
                happened_before = True
                continue  # want to get the metadata at the bottom of the file too
            if any(s in line for s in ("Start time", "End time", "Title")):
                tmp = line.strip("#").split(":")
                new_data.metadata[tmp[0].strip()] = ":".join(tmp[1:]).strip()
        column_headers = f[header_line].strip("#\t ").split("\t")
        data = np.genfromtxt(f, dtype="float", delimiter="\t", invalid_raise=False, comments="#")
    new_data.data = data[:, 0:-2]  # Deal with an errant tab at the end of each line
    new_data.column_headers = column_headers
    if np.all([h in column_headers for h in ("T (C)", "R (Ohm)")]):
        new_data.setas(x="T (C)", y="R (Ohm)")  # pylint: disable=not-callable
    return new_data



# -*- coding: utf-8 -*-
"""Provide some decorators and associated functions for modifying package classes and functions."""

from functools import wraps
import inspect
from importlib import import_module
from collections.abc import Iterable
from copy import copy
from os import environ

import numpy as np

from .tests import isproperty

_RTD = "READTHEDOCS" in environ


def image_file_adaptor(workingfunc):
    """Make wrappers for ImageFile functions.

    Notes:
        The wrapped functions take additional keyword arguments that are stripped off from the call.

    Keyword Arguments:
        _box(:py:meth:`Stoner.ImageArray.crop` arguments):
            Crops the image first before calling the parent method.
        _(bool, None):
            Controls whether a :py:class:`ImageArray` return will be substituted for the current
            :py:class:`ImageArray`.

            * True: - all ImageArray return types are substituted.
            * False (default) - Imagearray return types are substituted if they are the same size as the original
            * None - A copy of the current object is taken and the returned ImageArray provides the data.
    """
    # Avoid PEP257/black issue

    @wraps(workingfunc)
    def gen_func(self, *args, **kargs):
        """Wrap a called method to capture the result back into the calling object."""
        box = kargs.pop("_box", False)
        transpose = getattr(workingfunc, "transpose", False)
        if isinstance(box, bool) and not box:
            im = self.image
        else:
            im = self.image[im._box(box)]
        if transpose:
            im = im.T
        args = list(args)
        for ix, a in enumerate(args):
            if isinstance(a, type(self)):
                args[ix] = a.image

        if getattr(workingfunc, "changes_size", False) and "_" not in kargs:
            # special case for common function crop which will change the array shape
            force = True
        else:
            force = kargs.pop("_", False)
        r = workingfunc(im, *args, **kargs)
        if getattr(workingfunc, "keep_class", False):
            return r
        if isinstance(r, np.ndarray) and np.prod(r.shape) == np.max(r.shape):  # 1D Array
            ret = make_Data(r)
            ret.metadata = self.metadata.copy()
            ret.column_headers[0] = workingfunc.__name__
        elif isinstance(r, np.ndarray):  # make sure we return a ImageArray
            if transpose:
                r = r.T
            if isinstance(r, type(im)) and np.shares_memory(r, im):  # Assume everything was inplace
                self.image = r
                return self
            r = r.view(type(im))
            if r.shape == self.shape:
                if im.metadata is not r.metadata:
                    self.image = r.clone  # We're going to need to clone the return data because we can do fast copies.
                self.image[...] = r[...]
                self.metadata.update(r.metadata)
                return self
            ret = self.clone if not force else self
            ret.image = r.view(type(im))
            metadata = copy(self.metadata)
            metadata.update(r.metadata)
            ret.metadata = metadata
            return ret
        else:
            return r

    return fix_signature(gen_func, workingfunc)


def image_file_raw_adaptor(workingfunc):
    """Make wrappers for ImageFile functions.

    Notes:
        The wrapped functions take additional keyword arguments that are stripped off from the call.

    Keyword Arguments:
        _box(:py:meth:`Stoner.ImageArray.crop` arguments):
            Crops the image first before calling the parent method.
        _(bool, None):
            Controls whether a :py:class:`ImageArray` return will be substituted for the current
            :py:class:`ImageArray`.

            * True: - all ImageArray return types are substituted.
            * False (default) - Imagearray return types are substituted if they are the same size as the original
            * None - A copy of the current object is taken and the returned ImageArray provides the data.
    """
    # Avoid PEP257/black issue

    @wraps(workingfunc)
    def gen_func(self, *args, **kargs):
        """Wrap a called method to capture the result back into the calling object."""
        box = kargs.pop("_box", False)
        transpose = getattr(workingfunc, "transpose", False)
        clones = getattr(workingfunc, "clones", False)
        if isinstance(box, bool) and not box:
            im = self.image
        else:
            im = self.image[im._box(box)]
        if transpose:
            im = im.T
        args = list(args)
        for ix, a in enumerate(args):
            if isinstance(a, type(self)):
                args[ix] = a.image

        if getattr(workingfunc, "changes_size", False) and "_" not in kargs:
            # special case for common function crop which will change the array shape
            force = True
        else:
            force = kargs.pop("_", False)
        r = workingfunc(im, *args, **kargs)
        if getattr(workingfunc, "keep_class", False):
            return r
        if isinstance(r, np.ndarray) and r.ndim != 2:  # 1D Array goes back straight
            return r
        if isinstance(r, np.ndarray):  # make sure we return a ImageArray
            if transpose:
                r = r.T
            ret = self if not clones else self.clone
            if isinstance(r, type(im)) and np.shares_memory(r, im):  # Assume everything was inplace
                ret.image = r
                return ret
            r = r.view(type(im))
            if r.shape == ret.shape:
                ret.image = ret.image.astype(r.dtype)
                ret.image[...] = r[...]
                ret.metadata.update(r.metadata)
                return ret
            ret = self.clone if not force else self
            ret.image = r.view(type(im))
            metadata = copy(ret.metadata)
            metadata.update(r.metadata)
            ret.metadata = metadata
            return ret
        return r

    return fix_signature(gen_func, workingfunc)


def array_file_property(workingfunc):
    """Wrap an arbitrary callbable to make it a bound method of this class.

    Args:
        workingfunc (callable):
            The callable object to be wrapped.

    Returns:
        (function):
            A function with enclosure that holds additional information about this object.

    The function object returned simply calls the working function having got the _image property.
    """
    if workingfunc is None:  # We may not in fact be wrapping anything here!
        return None

    @wraps(workingfunc)
    def gen_func(self, *args, **kargs):
        """Wrap magic proxy function call."""
        transpose = getattr(workingfunc, "transpose", False)
        clones = getattr(workingfunc, "clones", False)
        im = self.image if not clones else self.clone.image
        if transpose:
            im = im.T
        args = list(args)
        for ix, a in enumerate(args):
            if isinstance(a, type(self)):
                args[ix] = a.image

        ret = workingfunc(im, *args, **kargs)
        # This shouldn't in fact be returning anything
        return ret

    return fix_signature(gen_func, workingfunc)


def array_file_attr(name):
    """Construct a property that will handle getting setting ande deleting the name attribute."""

    def getter(self):
        return getattr(self._image, name)

    def setter(self, value):
        return setattr(self._image, name, value)

    def deleter(self):
        return delattr(self._image, name)

    return property(getter, setter, deleter, f"Pass thrpough for {name}")


def image_array_adaptor(workingfunc):
    """Wrap an arbitrary callbable to make it a bound method of this class.

    Args:
        workingfunc (callable):
            The callable object to be wrapped.

    Returns:
        (function):
            A function with enclosure that holds additional information about this object.

    The function returned from here will call workingfunc with the first argument being a clone of this
    ImageArray. If the meothd returns an ndarray, it is wrapped back to our own class and the metadata dictionary
    is updated. If the function returns a :py:class:`Stoner.Data` object then this is also updated with our
    metadata.

    This method also updates the name and documentation strings for the wrapper to match the wrapped function -
    thus ensuring that Spyder's help window can generate useful information.

    """
    # Avoid PEP257/black issue

    @wraps(workingfunc)
    def gen_func(self, *args, **kwargs):
        """Wrap magic proxy function call."""
        transpose = getattr(workingfunc, "transpose", False)
        if transpose:
            change = self.T
        else:
            change = self
        r = workingfunc(change, *args, **kwargs)  # send copy of self as the first arg
        if isinstance(r, make_Data(None)):
            pass  # Data return is ok
        elif isinstance(r, np.ndarray) and np.prod(r.shape) == np.max(r.shape):  # 1D Array
            r = make_Data(r)
            r.metadata = self.metadata.copy()
            r.column_headers[0] = workingfunc.__name__
        elif isinstance(r, np.ndarray):  # make sure we return a ImageArray
            if transpose:
                r = r.T
            if isinstance(r, type(self)) and np.shares_memory(r, self):  # Assume everything was inplace
                return r
            r = r.view(type(self))
            sm = self.metadata.copy()  # Copy the currently metadata
            sm.update(r.metadata)  # merge in any new metadata from the call
            r.metadata = sm  # and put the returned metadata as the merged data
        # NB we might not be returning an ndarray at all here !
        return r

    gen_func.keep_class = getattr(workingfunc, "keep_class", False)
    return fix_signature(gen_func, workingfunc)


def class_modifier(
    module,
    adaptor=image_array_adaptor,
    transpose=False,
    overload=False,
    proxy_cls=None,
    RTD_restrictions=True,
    no_long_names=False,
):
    """Decorate  a class by addiding member functions from module.

    The purpose of this is to incorporate the functions within a module into being methods of the class being
    defined here.

    Args:
        cls (class):
            The class being defined
        module (imported module):
            The module whose functions members should be added to the class.

    Keyword Arguments:
        adaptor (callable):
            The factor function that takes the module function and produces a method that will call the function and
            take care of adapting the result.
        transpose (bool):
            Whether ther functions in the module need to have their data transposed to work.
        overload (bool):
            If False, don't overwrite the existing method.'
        proxy (class,None):
            If not None, the class whose attributes we are being augmented with these functions - need to check for
            clashing names.
        RTD_Restrictions (bool):
            If True (default), do not add members from outside our own package when on ReadTheDocs.
        no_long_names (bool):
            To avoid name collision the default is to create two entries in the class __dict__ - one for the standard name
            and one to include the full module path. This disables the latter.


    Returns:
        (class):
            The class with the additional methods added to it.
    """

    def actual_decorator(cls):
        proxy_class = cls if proxy_cls is None else proxy_cls
        mods = module if isinstance(module, Iterable) else [module]
        for mod in mods:
            if (RTD_restrictions and _RTD) and not getattr(mod, "__package__", "Stoner").startswith("Stoner"):
                continue  # Do not bind all the external functions if we're in ReadTheDocs
            for fname in dir(mod):
                if not fname.startswith("_"):
                    try:
                        func = getattr(mod, fname)
                    except AttributeError:  # This shouldn't happen, but it did for scipy.ndimage!
                        continue
                    fmod = getattr(func, "__module__", getattr(getattr(func, "__class__", None), "__module__", ""))
                    if callable(func) and isinstance(fmod, str) and fmod[:5] in ["Stone", "scipy", "skima"]:
                        if transpose:
                            func.transpose = transpose
                        name = f"{fmod}__{fname}".replace(".", "__")
                        proxy = adaptor(func)
                        setattr(proxy, "_src_mod", fmod)
                        if not no_long_names:
                            setattr(cls, name, proxy)
                        if overload or fname not in dir(proxy_class):
                            setattr(cls, fname, proxy)
        return cls

    return actual_decorator


def class_wrapper(
    target=None,
    adaptor=image_file_raw_adaptor,
    getter_adaptor=image_file_raw_adaptor,
    setter_adaptor=array_file_property,
    deleter_adaptor=array_file_property,
    attr_pass=array_file_attr,
    exclude_below=None,
):
    """Create entries in the current class for all attributes of klass that are not already defined.

    Keyword Arguments:
        target (type):
            The target class whose attributes we're going to link through to.
        adaptor (callable):
            A factory function to make methods to the the connection to the underlying attributes.

    Returns:
        class:
            Modified class definition.

    Note:
        We exclude attributes with which have the attribute _src_mod as these are being patched already.
    """

    def actual_decorator(cls):
        for name in dir(target):
            if name.startswith("_"):
                continue
            attr = getattr(target, name)
            if callable(attr) and not isproperty(target, name) and name not in dir(cls):
                proxy = adaptor(attr)
                setattr(cls, name, proxy)
            elif (
                isproperty(target, name)
                and hasattr(attr, "fget")
                and (name not in dir(cls) or name in dir(exclude_below))
            ):
                fget = getter_adaptor(getattr(attr, "fget"))
                fset = setter_adaptor(getattr(attr, "fset", None))
                fdel = deleter_adaptor(getattr(attr, "fdel", None))
                doc = getattr(attr, "__doc__", "")
                setattr(cls, name, property(fget, fset, fdel, doc))
            elif name not in cls.__dict__ and not callable(attr) and not isproperty(target, name):
                setattr(cls, name, attr_pass(name))
        return cls

    return actual_decorator


def changes_size(func):
    """Mark a function as one that changes the size of the ImageArray."""
    func.changes_size = True
    return func


def keep_return_type(func):
    """Mark a function as one that Should not be converted from an array to an ImageFile."""
    func.keep_class = True
    return func


def clones(func):
    """Mark the method as one that expects it's input to be cloned."""
    func.clones = True
    return func


def fix_signature(proxy_func, wrapped_func):
    """Update proxy_func to have a signature that matches the wrapped func."""
    try:
        proxy_func.__wrapped__.__signature__ = inspect.signature(wrapped_func)
    except (AttributeError, ValueError):  # Non-critical error
        try:
            proxy_func.__signature__ = inspect.signature(wrapped_func)
        except (AttributeError, ValueError):
            pass
    if hasattr(wrapped_func, "changes_size"):
        proxy_func.changes_size = wrapped_func.changes_size
    return proxy_func


def make_Data(*args, **kargs):
    """Return an instance of Stoner.Data passig through constructor arguments.

    Calling make_Data(None) is a special case to return the Data class ratther than an instance
    """
    if len(args) == 1 and args[0] is None:
        return import_module("Stoner.core.data").Data
    return import_module("Stoner.core.data").Data(*args, **kargs)



"""Use curve_fit to fit a plane to some data."""

# pylint: disable=invalid-name
from numpy.random import normal, seed
from numpy import linspace, meshgrid, column_stack, array
import matplotlib.cm as cmap

from Stoner import Data

seed(12345)  # Ensire consistent random numbers!


def plane(coord, a, b, c):
    """Function to define a plane."""
    x, y = coord.T
    return c - (x * a + y * b)


coeefs = [1, -0.5, -1]
col = linspace(-10, 10, 8)
X, Y = meshgrid(col, col)
Z = plane(array([X, Y]).T, *coeefs) + normal(size=X.shape, scale=2.0)
d = Data(
    column_stack((X.ravel(), Y.ravel(), Z.ravel())),
    filename="Fitting a Plane",
    setas="xyz",
)

d.column_headers = ["X", "Y", "Z"]
d.plot_xyz(plotter="scatter", title=None, griddata=False, color="k")

popt, pcov = d.curve_fit(plane, [0, 1], 2, result=True)
col = linspace(-10, 10, 128)
X, Y = meshgrid(col, col)
Z = plane(array([X, Y]).T, *popt)
e = Data(
    column_stack((X.ravel(), Y.ravel(), Z.ravel())),
    filename="Fitting a Plane",
    setas="xyz",
)
e.column_headers = d.column_headers

e.plot_xyz(linewidth=0, cmap=cmap.jet, alpha=0.5, figure=d.fig)

txt = "$z=c-ax+by$\n"
txt += "\n".join(
    [d.format("plane:{}".format(k), latex=True) for k in ["a", "b", "c"]]
)

ax = d.axes[0]
ax.text(-30, -10, 10, txt)



# -*- coding: utf-8 -*-
"""Provides classes and functions to support the :py:attr:`Stoner.DataFolder.metadata` magic attribute."""

__all__ = ["MetadataProxy"]
import fnmatch
from collections.abc import MutableMapping

from lmfit import Model
import numpy as np

from ..core import typeHintedDict, metadataObject
from ..compat import string_types
from ..tools import isLikeList, isiterable, make_Data
from ..Core import DataFile


def _fmt_as_list(results):
    """Convert the results list of slicing to a simple list."""
    keys = set()
    for r in results:
        keys |= set(r.keys())
    keys = list(keys)
    if len(keys) == 1:
        ret = [r.get(keys[0], None) for r in results]
    else:
        ret = []
        for r in results:
            ret.append(tuple(r.get(k, None) for k in keys))

    return ret


def _fmt_as_dict(results):
    """Non-opt, results is already adictionary."""
    return results


def _fmt_as_dataframe(results):
    """Format the return results as a DataFrame."""
    from pandas import DataFrame

    frame = DataFrame(results)
    return frame


def _fmt_as_Data(results):
    """Format the results as a Data() object."""
    ret = make_Data(_fmt_as_dataframe(results))
    mask = np.zeros(ret.shape, dtype=bool)
    for ix, col in enumerate(ret.data.T):
        try:
            mask[:, ix] = np.isnan(col)
        except TypeError:
            pass
    ret.mask = mask
    return ret


def _fmt_as_array(results):
    """Format the results as an array."""
    ret = _fmt_as_Data(results)
    if ret.data.shape[1] != 1:
        ret = ret.data
    else:
        ret = ret.data[:, 0]
    return ret


def _fmt_as_smart(results):
    """Decide what formatting of results to do."""
    if np.all([len(r) == 1 and list(r.keys())[0] == list(results[0].keys())[0] for r in results]):
        return _fmt_as_list(results)
    return _fmt_as_dict(results)


def _slice_keys(args, possible=None):
    """Work through the arguments to slice() and construct a list of keys."""
    keys = []
    for k in args:
        if isinstance(k, string_types):
            if k not in possible:
                sub_k = fnmatch.filter(possible, k)
                if len(sub_k) > 0:
                    keys.extend(_slice_keys(sub_k, possible))
                else:
                    raise KeyError(f"No matching keys for {sub_k} in metadata")
            else:
                keys.append(k)
        elif isinstance(k, type) and issubclass(k, Model):
            model = k.__name__
            for name in k().param_names:
                for sub_k in [f"{model}:{name}", f"{model}:{name} err"]:
                    if sub_k not in possible:
                        raise KeyError(f"No matching keys for {sub_k} in metadata")
                    keys.append(sub_k)
        elif isinstance(k, Model):
            model = type(k).__name__
            for name in k.param_names:
                for sub_k in [f"{model}:{name}", f"{model}:{name} err"]:
                    if sub_k not in possible:
                        raise KeyError(f"No matching keys for {sub_k} in metadata")
                    keys.append(sub_k)
        elif isiterable(k):
            keys.extend(_slice_keys(k, possible))
        else:
            raise KeyError(f"{type(k)} cannot be used as a key name or set of key names")
    return keys


class MetadataProxy(MutableMapping):
    """Provide methods to interact with a whole collection of metadataObjects' metadata."""

    def __init__(self, folder):
        """Note our parent folder object."""
        self._folder = folder

    @property
    def all(self):
        """List all the metadata dictionaries in the Folder."""
        if hasattr(self._folder, "_metadata"):  # Extra logic for Folders like Stack
            for item in self._folder._metadata.items():
                yield item
        else:
            for item in self._folder:
                yield item.metadata

    @all.setter
    def all(self, value):
        """List all the metadata dictionaries in the Folder."""
        if hasattr(self._folder, "_metadata"):  # Direct support for metadata dictionary
            for new, old in zip(value, self._folder._metadata):
                old.update(new)
        else:
            for new, item in zip(value, self._folder):
                item.metadata.update(new)

    @property
    def all_by_keys(self):
        """Return the set of metadata keys common to all objects int he Folder."""
        if len(self._folder) > 0:
            keys = set(self._folder[0].metadata.keys())
            for d in self._folder:
                keys &= set(d.metadata.keys())
        else:
            keys = set()
        ret = typeHintedDict()
        for k in sorted(list(keys)):
            ret[k] = self[k].view(np.ndarray)
        return ret

    @property
    def common_keys(self):
        """Return the set of metadata keys common to all objects int he Folder."""
        if len(self._folder) > 0:
            keys = set(self._folder[0].metadata.keys())
            for d in self._folder:
                keys &= set(d.metadata.keys())
        else:
            keys = set()
        return sorted(list(keys))

    @property
    def common_metadata(self):
        """Return a dictionary of the common_keys that have common values."""
        output = typeHintedDict()
        for key in self.common_keys:
            vals = self.slice(key, output="list")
            if np.all(vals == vals[0]):
                output[key] = vals[0]
        return output

    def __contains__(self, item):
        """Check for membership of all possible kes."""
        return item in self.all_keys()

    def __iter__(self):
        """Iterate over objects."""
        for k in self.common_keys:
            yield k

    def __len__(self):
        """Out length is our common_keys."""
        return len(self.common_keys)

    def __repr__(self):
        """Give an informative display of the metadata representation."""
        return (
            f"The {type(self._folder).__name__} {self._folder.key} has"
            + f" {len(self)} common keys of metadata in {len(self._folder)} {self._folder.type.__name__} objects"
        )

    def __delitem__(self, item):
        """Attempt to delete item from all members of the folder."""
        ok = False
        for entry in self._folder:
            try:
                del entry.metadata[item]
                ok = True
            except KeyError:
                pass
        if not ok:  # item was not a key in any data file
            raise KeyError(f"{item} was not recognised as a metadata key in any object in the folder.")

    def __getitem__(self, value):
        """Return an array formed by getting a single key from each object in the Folder."""
        ret = self.slice(value, mask_missing=True, output="array")
        if ret.size == 0:
            raise KeyError(f"{value} did not match any keys in any file")
        return ret

    def __setitem__(self, key, value):
        """Proxy to set an item on all the entries in the folder."""
        for d in self._folder:
            d[key] = value

    def __xor__(self, other):
        """Implement an XOR operator that gives differences between metadata dictionaries."""
        if isinstance(other, type(self._folder)):
            other = other.metadata
        if isinstance(other, MetadataProxy):
            other = other.all_by_keys
        elif isinstance(other, metadataObject):
            other = other.metadata
        else:
            return NotImplemented
        return self.all_by_keys ^ other

    def __eq__(self, other):
        """Equality test operator."""
        ret = self ^ other
        if not isinstance(ret, dict):
            return NotImplemented
        return len(ret) == 0

    def all_keys(self):
        """Return the union of all the metadata keyus for all objects int he Folder."""
        if len(self._folder) > 0:
            keys = set(self._folder[0].metadata.keys())
            for d in self._folder:
                keys |= set(d.metadata.keys())
        else:
            keys = set()
        for k in sorted(keys):
            yield k

    def all_items(self):
        """Return the result of indexing the metadata with all_keys().

        Yields:
            key,self[key]
        """
        for k in self.all_keys():
            yield k, self[k]

    def all_values(self):
        """Return the result of indexing the metadata with all_keys().

        Yields:
            self[key]
        """
        for k in self.all_keys():
            yield self[k]

    def apply(self, key, func):
        """Evaluate a function for each item in the folder and store the return value in a metadata key.

        Args:
            key (str): The name of the key to store the result in.
            func(callable): The function to be evaluated.

        Returns:
            (self) a copy of the combined metadata object to allow routines to be strung together.

        Notes:
            The function should have  a protoptye of the form:

                def func(i,metadataObject):

            where i is a counter that runs from 0 to the length of the current Folder
            and metadataObject will be each object in the Folder in turn.
        """
        for i, d in enumerate(self._folder):
            d[key] = func(i, d)
        return self

    def slice(self, *args, **kwargs):  # pylint: disable=arguments-differ
        """Return a list of the metadata dictionaries for each item/file in the top level group.

        Keyword Arguments:
            *args (string, lmfit.Model class or instance  or iterable of string, lmfit Models):
                if given then only return the item(s) requested from the metadata
            values_only(bool):
                if given and *output* not set only return tuples of the dictionary values. Mostly useful
                when given a single key string
            output (str or type):
                Controls the output format from slice_metadata. Possible values are

                - "dict" or dict - return a list of dictionary subsets of the metadata from each image
                - "list" or list - return a list of values of each item pf the metadata
                - "array" or np.array - return a single array - like list above, but returns as a numpy array.
                  This can create a 2D array from multiple keys
                - "data" or Stoner.Data - returns the metadata in a Stoner.Data object where the column headers
                  are the metadata keys.
                - "frame" - returns the metadata as a Pandas DataFrame object
                - "smart" - switch between *dict* and *list* depending whether there is one or more keys.
            mask_missing (bool):
                If true, then metadata entries missing in members of the folder are returned as masked values (or
                None), If False, then an exception is raised if any entries are missing.

        Returns:
            ret(list of dict, tuple of values or :py:class:`Stoner.Data`):
                depending on *values_only* or (output* returns the sliced dictionaries or tuples/
                values of the items

        """
        values_only = kwargs.pop("values_only", False)
        output = kwargs.pop("output", None)
        mask_missing = kwargs.pop("mask_missing", False)
        if kwargs:
            raise SyntaxError(f"Unused keyword arguments : {kwargs}")
        if output is None:  # Sort out a definitive value of output
            output = "dict" if not values_only else "smart"
        if isinstance(output, string_types):
            output = output.lower()
        outputs = {
            "list": _fmt_as_list,
            list: _fmt_as_list,
            "dict": _fmt_as_dict,
            dict: _fmt_as_dict,
            "frame": _fmt_as_dataframe,
            "data": _fmt_as_Data,
            DataFile: _fmt_as_Data,
            "array": _fmt_as_array,
            np.ndarray: _fmt_as_array,
            "smart": _fmt_as_smart,
        }
        if output not in outputs:  # Check for good output value
            raise TypeError(f"output of slice metadata must be either dict, list, or array not {output}")
        formatter = outputs[output]
        possible = list(self.all_keys()) if mask_missing else self.common_keys
        keys = _slice_keys(args, possible)
        results = []
        for d in self._folder:
            results.append({k: d[k] for k in keys if k in d})

        for r in results:  # Expand the results where a result contains a list
            for k in keys:
                if k in r and isLikeList(r[k]) and len(r[k]) > 0:
                    v = r[k]
                    del r[k]
                    r.update({f"{k}[{i}]": vi for i, vi in enumerate(v)})

        return formatter(results)



"""Simple use of lmfit to fit data."""

# pylint: disable=invalid-name
from numpy import linspace, exp, random

from Stoner import Data
from Stoner.plot.utils import errorfill

random.seed(12345)  # Ensure consistent random numbers!

# Make some data
x = linspace(0, 10.0, 101)
y = 2 + 4 * exp(-x / 1.7) + random.normal(scale=0.2, size=101)
x += +random.normal(scale=0.1, size=101)

d = Data(x, y, column_headers=["Time", "Signal"], setas="xy")

func = lambda x, A, B, C: A + B * exp(-x / C)

# Do the fitting and plot the result
fit = d.odr(
    func,
    result=True,
    header="Fit",
    A=1,
    B=1,
    C=1,
    prefix="Model",
    residuals=True,
)

# Reset labels
d.labels = []

# Make nice two panel plot layout
ax = d.subplot2grid((3, 1), (2, 0))
d.setas = "x..y"
d.plot(fmt="g+")
d.title = ""

# Plot up the data
ax = d.subplot2grid((3, 1), (0, 0), rowspan=2)
d.setas = "xy"
d.plot(fmt="ro")

d.setas = "x.y"
d.plot(plotter=errorfill, yerr=0.2, color="orange")
d.plot(plotter=errorfill, xerr=0.1, color="orange", label=None)
d.xticklabels = [[]]
d.xlabel = ""

# Annotate plot with fitting parameters
d.annotate_fit(func, prefix="Model", x=0.7, y=0.3, fontdict={"size": "x-small"})
text = r"$y=A+Be^{-x/C}$" + "\n\n"
d.text(7.2, 3.9, text, fontdict={"size": "x-small"})
d.title = "Orthogonal Distance Regression  Fit"



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Channel column operations functions for analysis code."""

__all__ = ["ColumnOpsMixin"]

import numpy as np

from Stoner.tools import isiterable, all_type
from Stoner.compat import index_types


class ColumnOpsMixin:
    """A mixin calss designed to work with :py:class:`Stoner.Core.DataFile` to provide additional stats methods."""

    def _do_error_calc(self, col_a, col_b, error_type="relative"):
        """Do an error calculation."""
        col_a = self.find_col(col_a)
        error_calc = None
        if (
            isinstance(col_a, (list, tuple))
            and isinstance(col_b, (list, tuple))
            and len(col_a) == 2
            and len(col_b) == 2
        ):  # Error columns on
            (col_a, e1) = col_a
            (col_b, e2) = col_b
            e1data = self.__get_math_val(e1)[0]
            e2data = self.__get_math_val(e2)[0]
            if error_type == "relative":

                def error_calc(adata, bdata):  # pylint: disable=function-redefined
                    """Relative error summation."""
                    return np.sqrt((e1data / adata) ** 2 + (e2data / bdata) ** 2)

            elif error_type == "absolute":

                def error_calc(adata, bdata):  # pylint: disable=function-redefined, unused-argument
                    """Sum absolute errors."""
                    return np.sqrt(e1data**2 + e2data**2)

            elif error_type == "diffsum":

                def error_calc(adata, bdata):  # pylint: disable=function-redefined
                    """Calculate error for difference over sum."""
                    return np.sqrt(
                        (1.0 / (adata + bdata) - (adata - bdata) / (adata + bdata) ** 2) ** 2 * e1data**2
                        + (-1.0 / (adata + bdata) - (adata - bdata) / (adata + bdata) ** 2) ** 2 * e2data**2
                    )

            else:
                raise ValueError(f"Unknown error calculation mode {error_type}")

        adata, aname = self.__get_math_val(col_a)
        bdata, bname = self.__get_math_val(col_b)
        return adata, bdata, error_calc, aname, bname

    def __get_math_val(self, col):
        """Interpret col as either col_a column index or value or an array of values.

        Args:
            col (various):
                If col can be interpreted as col_a column index then return the first matching column.
                If col is col_a 1D array of the same length as the data then just return the data. If col is col_a
                float then just return it as col_a float.

        Returns:
            (tuple of (:py:class:`Stoner.cpre.DataArray`,str)):
                The matching data.
        """
        if isinstance(col, index_types):
            col = self.find_col(col)
            if isinstance(col, list):
                col = col[0]
            data = self.column(col)
            name = self.column_headers[col]
        elif isinstance(col, np.ndarray) and col.ndim == 1 and len(col) == len(self):
            data = col
            name = "data"
        elif isinstance(col, float):
            data = col * np.ones(len(self))
            name = str(col)
        else:
            raise RuntimeError(f"Bad column index: {col}")
        return data, name

    def add(self, col_a, col_b, replace=False, header=None, index=None):
        """Add one column, number or array (col_b) to another column (col_a).

        Args:
            col_a (index):
                First column to work with
            col_b (index, float or 1D array):
                Second column to work with.

        Keyword Arguments:
            header (string or None):
                new column header  (defaults to a-b
            replace (bool):
                Replace the col_a column with the new data
            index (column index or None):
                Column to insert new data at.

        Returns:
            (:py:class:`Stoner.Data`):
                The newly modified Data object.

        If col_a and col_b are tuples of length two, then the firstelement is assumed to be the value and
        the second element an uncertainty in the value. The uncertainties will then be propagated and an
        additional column with the uncertainites will be added to the data.
        """
        adata, bdata, err_calc, aname, bname = self._do_error_calc(col_a, col_b, error_type="absolute")
        err_header = None
        if isinstance(header, tuple) and len(header) == 2:
            header, err_header = header
        if header is None:
            header = f"{aname}+{bname}"
        if err_calc is not None and err_header is None:
            err_header = "Error in " + header
        index = self.shape[1] if index is None else self.find_col(index)
        if err_calc is not None:
            err_data = err_calc(adata, bdata)
        self.add_column((adata + bdata), header=header, index=index, replace=replace)
        if err_calc is not None:
            self.add_column(err_data, header=err_header, index=index + 1, replace=False)
        return self

    def diffsum(self, col_a, col_b, replace=False, header=None, index=None):
        r"""Calculate :math:`\frac{a-b}{a+b}` for the two columns *a* and *b*.

        Args:
            col_a (index):
                First column to work with
            col_b (index, float or 1D array):
                Second column to work with.

        Keyword Arguments:
            header (string or None):
                new column header  (defaults to a-b
            replace (bool):
                Replace the col_a column with the new data
            index (column index or None):
                Column to insert new data at.

        Returns:
            (:py:class:`Stoner.Data`):
                The newly modified Data object.

        If col_a and col_b are tuples of length two, then the firstelement is assumed to be the value and
        the second element an uncertainty in the value. The uncertainties will then be propagated and an
        additional column with the uncertainites will be added to the data.
        """
        adata, bdata, err_calc, aname, bname = self._do_error_calc(col_a, col_b, error_type="diffsum")
        err_header = None
        if isinstance(header, tuple) and len(header) == 2:
            header, err_header = header
        if header is None:
            header = f"({aname}-{bname})/({aname}+{bname})"
        if err_calc is not None and err_header is None:
            err_header = "Error in " + header
        if err_calc is not None:
            err_data = err_calc(adata, bdata)
        index = self.shape[1] if index is None else self.find_col(index)
        self.add_column((adata - bdata) / (adata + bdata), header=header, index=index, replace=replace)
        if err_calc is not None:
            self.add_column(err_data, header=err_header, index=index + 1, replace=False)
        return self

    def divide(self, col_a, col_b, replace=False, header=None, index=None):
        """Divide one column (col_a) by  another column, number or array (col_b).

        Args:
            col_a (index):
                First column to work with
            col_b (index, float or 1D array):
                Second column to work with.

        Keyword Arguments:
            header (string or None):
                new column header  (defaults to a-b
            replace (bool):
                Replace the col_a column with the new data
            index (column index or None):
                Column to insert new data at.

        Returns:
            (:py:class:`Stoner.Data`):
                The newly modified Data object.

        If col_a and col_b are tuples of length two, then the firstelement is assumed to be the value and
        the second element an uncertainty in the value. The uncertainties will then be propagated and an
        additional column with the uncertainites will be added to the data.
        """
        adata, bdata, err_calc, aname, bname = self._do_error_calc(col_a, col_b, error_type="relative")
        err_header = None
        if isinstance(header, tuple) and len(header) == 2:
            header, err_header = header
        if header is None:
            header = f"{aname}/{bname}"
        if err_calc is not None and err_header is None:
            err_header = "Error in " + header
        index = self.shape[1] if index is None else self.find_col(index)
        self.add_column((adata / bdata), header=header, index=index, replace=replace)
        if err_calc is not None:
            err_data = err_calc(adata, bdata) * np.abs(adata / bdata)
            self.add_column(err_data, header=err_header, index=index + 1, replace=False)
        return self

    def max(self, column=None, bounds=None):
        """Find maximum value and index in col_a column of data.

        Args:
            column (index):
                Column to look for the maximum in

        Keyword Arguments:
            bounds (callable):
                col_a callable function that takes col_a single argument list of
                numbers representing one row, and returns True for all rows to search in.

        Returns:
            (float,int):
                (maximum value,row index of max value)

        Note:
            If column is not defined (or is None) the :py:attr:`DataFile.setas` column
            assignments are used.
        """
        if column is None:
            col = self.setas._get_cols("ycol")
        else:
            col = self.find_col(column)
        if bounds is not None:
            self._push_mask()
            self._set_mask(bounds, True, col)
        result = self.data[:, col].max(), self.data[:, col].argmax()
        if bounds is not None:
            self._pop_mask()
        return result

    def mean(self, column=None, sigma=None, bounds=None):
        """Find mean value of col_a data column.

        Args:
            column (index):
                Column to look for the maximum in

        Keyword Arguments:
            sigma (column index or array):
                The uncertainty noted for each value in the mean
            bounds (callable):
                col_a callable function that takes col_a single argument list of
                numbers representing one row, and returns True for all rows to search in.

        Returns:
            (float):
                The mean of the data.

        Note:
            If column is not defined (or is None) the :py:attr:`DataFile.setas` column
            assignments are used.

        .. todo::
            Fix the row index when the bounds function is used - see note of :py:meth:`AnalysisMixin.max`
        """
        _ = self._col_args(scalar=True, ycol=column, yerr=sigma)

        if bounds is not None:
            self._push_mask()
            self._set_mask(bounds, True, _.ycol)

        if isiterable(sigma) and len(sigma) == len(self) and all_type(sigma, float):
            sigma = np.array(sigma)
            _["has_yerr"] = True
        elif _.has_yerr:
            sigma = self.data[:, _.yerr]

        if not _.has_yerr:
            result = self.data[:, _.ycol].mean()
        else:
            ydata = self.data[:, _.ycol]
            w = 1 / (sigma**2 + 1e-8)
            norm = w.sum(axis=0)
            error = np.sqrt((sigma**2).sum(axis=0)) / len(sigma)
            result = (ydata * w).mean(axis=0) / norm, error
        if bounds is not None:
            self._pop_mask()
        return result

    def min(self, column=None, bounds=None):
        """Find minimum value and index in col_a column of data.

        Args:
            column (index):
                Column to look for the maximum in

        Keyword Arguments:
            bounds (callable):
                col_a callable function that takes col_a single argument list of
                numbers representing one row, and returns True for all rows to search in.

        Returns:
            (float,int):
                (minimum value,row index of min value)

        Note:
            If column is not defined (or is None) the :py:attr:`DataFile.setas` column
            assignments are used.
        """
        if column is None:
            col = self.setas._get_cols("ycol")
        else:
            col = self.find_col(column)
        if bounds is not None:
            self._push_mask()
            self._set_mask(bounds, True, col)
        result = self.data[:, col].min(), self.data[:, col].argmin()
        if bounds is not None:
            self._pop_mask()
        return result

    def multiply(self, col_a, col_b, replace=False, header=None, index=None):
        """Multiply one column (col_a) by  another column, number or array (col_b).

        Args:
            col_a (index):
                First column to work with
            col_b (index, float or 1D array):
                Second column to work with.

        Keyword Arguments:
            header (string or None):
                new column header  (defaults to a-b
            replace (bool):
                Replace the col_a column with the new data
            index (column index or None):
                Column to insert new data at.

        Returns:
            (:py:class:`Stoner.Data`):
                The newly modified Data object.

        If col_a and col_b are tuples of length two, then the firstelement is assumed to be the value and
        the second element an uncertainty in the value. The uncertainties will then be propagated and an
        additional column with the uncertainites will be added to the data.
        """
        adata, bdata, err_calc, aname, bname = self._do_error_calc(col_a, col_b, error_type="relative")
        err_header = None
        if isinstance(header, tuple) and len(header) == 2:
            header, err_header = header
        if header is None:
            header = f"{aname}*{bname}"
        if err_calc is not None and err_header is None:
            err_header = "Error in " + header
        index = self.shape[1] if index is None else self.find_col(index)
        if err_calc is not None:
            err_data = err_calc(adata, bdata) * np.abs(adata * bdata)
        self.add_column((adata * bdata), header=header, index=index, replace=replace)
        if err_calc is not None:
            self.add_column(err_data, header=err_header, index=index + 1, replace=False)
        return self

    def span(self, column=None, bounds=None):
        """Return a tuple of the maximum and minimum values within the given column and bounds.

        Args:
            column (index):
                Column to look for the maximum in

        Keyword Arguments:
            bounds (callable):
                col_a callable function that takes col_a single argument list of
                numbers representing one row, and returns True for all rows to search in.

        Returns:
            (float,float):
                col_a tuple of (min value, max value)

        Note:
            This works by calling into :py:meth:`Data.max` and :py:meth:`Data.min`.

            If column is not defined (or is None) the :py:attr:`DataFile.setas` column
            assignments are used.

        """
        return (self.min(column, bounds)[0], self.max(column, bounds)[0])

    def std(self, column=None, sigma=None, bounds=None):
        """Find standard deviation value of col_a data column.

        Args:
            column (index):
                Column to look for the maximum in

        Keyword Arguments:
            sigma (column index or array):
                The uncertainty noted for each value in the mean
            bounds (callable):
                col_a callable function that takes col_a single argument list of
                numbers representing one row, and returns True for all rows to search in.

        Returns:
            (float):
                The standard deviation of the data.

        Note:
            If column is not defined (or is None) the :py:attr:`DataFile.setas` column
            assignments are used.

        .. todo::
            Fix the row index when the bounds function is used - see note of :py:meth:`AnalysisMixin.max`
        """
        _ = self._col_args(scalar=True, ycol=column, yerr=sigma)

        if bounds is not None:
            self._push_mask()
            self._set_mask(bounds, True, _.ycol)

        if isiterable(sigma) and len(sigma) == len(self) and all_type(sigma, float):
            sigma = np.array(sigma)
        elif _.yerr:
            sigma = self.data[:, _.yerr]
        else:
            sigma = np.ones(len(self))

        ydata = self.data[:, _.ycol]

        sigma = np.abs(sigma) / np.nanmax(np.abs(sigma))
        sigma = np.where(sigma < 1e-8, 1e-8, sigma)
        weights = 1 / sigma**2
        weights[np.isnan(weights)] = 0.0

        result = np.sqrt(np.cov(ydata, aweights=weights))

        if bounds is not None:
            self._pop_mask()
        return result

    def subtract(self, col_a, col_b, replace=False, header=None, index=None):
        """Subtract one column, number or array (col_b) from another column (col_a).

        Args:
            col_a (index):
                First column to work with
            col_b (index, float or 1D array):
                Second column to work with.

        Keyword Arguments:
            header (string or None):
                new column header  (defaults to a-b
            replace (bool):
                Replace the col_a column with the new data
            index (column index or None):
                Column to insert new data at.

        Returns:
            (:py:class:`Stoner.Data`):
                The newly modified Data object.

        If col_a and col_b are tuples of length two, then the firstelement is assumed to be the value and
        the second element an uncertainty in the value. The uncertainties will then be propagated and an
        additional column with the uncertainites will be added to the data.
        """
        adata, bdata, err_calc, aname, bname = self._do_error_calc(col_a, col_b, error_type="absolute")
        err_header = None
        if isinstance(header, tuple) and len(header) == 2:
            header, err_header = header
        if header is None:
            header = f"{aname}-{bname}"
        if err_calc is not None and err_header is None:
            err_header = "Error in " + header
        index = self.shape[1] if index is None else self.find_col(index)
        if err_calc is not None:
            err_data = err_calc(adata, bdata)
        self.add_column((adata - bdata), header=header, index=index, replace=replace)
        if err_calc is not None:
            col_a = self.find_col(col_a)
            self.add_column(err_data, header=err_header, index=index + 1, replace=False)
        return self



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""The main Data Class definition."""

from sys import float_info
from Stoner.Analysis import AnalysisMixin
from Stoner.analysis.fitting.mixins import FittingMixin
from Stoner.analysis.columns import ColumnOpsMixin
from Stoner.analysis.filtering import FilteringOpsMixin
from Stoner.analysis.features import FeatureOpsMixin
from Stoner.plot import PlotMixin
from Stoner.Core import DataFile

# Bring all the subclasses into memory (idnore unused imports warnings)
import Stoner.formats  # NOQA pylint: disable=W0611
import Stoner.HDF5  # NOQA pylint: disable=W0611
import Stoner.Zip  # NOQA pylint: disable=W0611

from Stoner.tools import format_error


class Data(AnalysisMixin, FittingMixin, ColumnOpsMixin, FilteringOpsMixin, FeatureOpsMixin, PlotMixin, DataFile):
    """The principal class for representing a data file.

    This merges:
        -   :py:class:`Stoner.Core.DataFile`,
        -   :py:class:`Stoner.Analysis.AnalysisMixin`,
        -   :py:class:`Stoner.analysis.fitting.FittingMixin`,
        -   :py:class:`Stoner.analysis.columns.ColumnOpsMixin`,
        -   :py:class:`Stoner.analysis.filtering.FilteringOpsMixin`,
        -   :py:class:`Stoner.analysis.features.FeatureOpsMixin`,
        -   :py:class:`Stoner.plot.PlotMixin`

    Also has the :py:mod:`Stoner.formats` loaded redy for use.
    """

    def format(self, key, **kargs):
        r"""Return the contents of key pretty formatted using :py:func:`format_error`.

        Args:
            fmt (str): Specify the output format, options are:

                *  "text" - plain text output
                * "latex" - latex output
                * "html" - html entities

            escape (bool):
                Specifies whether to escape the prefix and units for unprintable characters in non
                text formats )default False)
            mode (string):
                If "float" (default) the number is formatted as is, if "eng" the value and error is converted
                to the next samllest power of 1000 and the appropriate SI index appended. If mode is "sci" then a
                scientific, i.e. mantissa and exponent format is used.
            units (string):
                A suffix providing the units of the value. If si mode is used, then appropriate si
                prefixes are prepended to the units string. In LaTeX mode, the units string is embedded in \mathrm
            prefix (string):
                A prefix string that should be included before the value and error string. in LaTeX mode this is
                inside the math-mode markers, but not embedded in \mathrm.

        Returns:
            A pretty string representation.

        The if key="key", then the value is self["key"], the error is self["key err"], the default prefix is
        self["key label"]+"=" or "key=", the units are self["key units"] or "".

        """
        mode = kargs.pop("mode", "float")
        units = kargs.pop("units", self.get(key + " units", ""))
        prefix = kargs.pop("prefix", f"{self.get(key + ' label', f'{key}')} = ")
        latex = kargs.pop("latex", False)
        fmt = kargs.pop("fmt", "latex" if latex else "text")
        escape = kargs.pop("escape", False)

        try:
            value = float(self[key])
        except (ValueError, TypeError) as err:
            raise KeyError(f"{key} should be a floating point value of the metadata not a {type(self[key])}.") from err
        try:
            error = float(self[f"{key} err"])
        except (TypeError, KeyError):
            error = float_info.epsilon
        return format_error(value, error, fmt=fmt, mode=mode, units=units, prefix=prefix, scape=escape)



# -*- coding: utf-8 -*-
"""Image Loading routines for HDF5 format."""
from copy import deepcopy

import numpy as np

from ..decorators import register_loader
from ...tools import make_Data


@register_loader(
    patterns=[(".hdf5", 16), (".hdf", 16)],
    mime_types=[("application/x-hdf", 16), ("application/x-hdf5", 16)],
    name="STXMImage",
    what="Image",
)
def load_stxm_image(new_data, filename, *args, **kargs):
    """Initialise and load a STXM image produced by Pollux.

    Keyword Args:
        regrid (bool):
            If set True, the gridimage() method is automatically called to re-grid the image to known coordinates.
    """
    regrid = kargs.pop("regrid", False)
    kargs.setdefault("filetype", "SLS_STXMFile")
    bcn = kargs.pop("bcn", False)
    d = make_Data(filename, *args, **kargs)
    new_data.image = d.data
    new_data.metadata = deepcopy(d.metadata)
    new_data.filename = d.filename
    if isinstance(regrid, tuple):
        new_data.gridimage(*regrid)
    elif isinstance(regrid, dict):
        new_data.gridimage(**regrid)
    elif regrid:
        new_data.gridimage()
    if bcn:
        if regrid:
            new_data.metadata["beam current"] = new_data.metadata["beam current"].gridimage()
        new_data.image /= new_data["beam current"]
    new_data.polarization = np.sign(new_data.get("collection.polarization.value", 0))
    return new_data



# -*- coding: utf-8 -*-
"""
test_Core.py
Created on Tue Jan 07 22:05:55 2014

@author: phygbu
"""

import sys
import os.path as path
import pytest
import numpy as np
from numpy import all

pth = path.dirname(__file__)
pth = path.realpath(path.join(pth, "../../../"))
sys.path.insert(0, pth)

from Stoner import Data, __home__

datadir = path.join(pth, "sample-data")
selfd = Data(path.join(path.dirname(__file__), "..", "CoreTest.dat"), setas="xy")


def test_setas_basics():
    selfd4 = Data(path.join(__home__, "..", "sample-data", "Cu_resistivity_vs_T.txt"))
    selfd4.setas = "2.y.x3."
    assert selfd4.setas.x == 4, "Failed to set set as from string with numbers."
    tmp = list(selfd4.setas)
    selfd4.setas = "..y.x..."
    assert list(selfd4.setas) == tmp, "Explicit expansion of setas not the same as with numbers."
    selfd4.setas(x="T (K)")
    selfd4.setas(y="rho", reset=False)
    assert list(selfd4.setas) == tmp, "setas from calls with and without reset failed"
    s = selfd4.setas.clone
    selfd4.setas = []
    assert list(selfd4.setas) == ["." for i in range(8)], "Failed to clear setas"
    selfd4.setas(s)
    assert list(selfd4.setas) == tmp, "setas from call with setas failed"
    selfd4.setas(reset="Yes")
    assert list(selfd4.setas) == ["." for i in range(8)], "Failed to clear setas"
    assert selfd4.setas._size == 8, "Size attribute failed."
    assert selfd4[0, :].setas._size == 8, "Size attribute for array row failed."
    assert selfd4[:, 0].setas._size == 1, "Size attribute for array column  failed."
    selfd4.column_headers[-3:] = ["Column", "Column", "Column"]
    assert selfd4.setas._unique_headers == [
        "Voltage",
        "Current",
        "$\\rho$ ($\\Omega$m)",
        "Resistance",
        "T (K)",
        "Column",
        6,
        7,
    ], "Unique Headers failed in setas."
    selfd4.setas("2.y.x3.")
    s = selfd4.setas.clone
    selfd4.setas.clear()
    selfd4.setas(s)
    assert selfd4.setas == "..y.x...", "setas set by call with setas argument failed"
    selfd4.setas(selfd4.setas)
    assert selfd4.setas == "..y.x...", "setas set by call with self argument failed"
    selfd4.setas()
    assert selfd4.setas == "..y.x...", "setas set by call with no argument failed"

    selfd4.setas -= ["x", "y"]
    assert selfd4.setas == "8.", "setas __sub__ with iterable failed"

    selfd4.setas("2.y.x3.")
    assert selfd4.setas == "..y.x...", "Equality test by string failed."
    assert selfd4.setas == "2.y.x3.", "Equality test by numbered string failed."
    assert selfd4.setas.to_string() == "..y.x...", "To_string() failed."
    assert selfd4.setas.to_string(encode=True) == "..y.x3.", "To_string() failed."
    selfd4.setas.clear()
    selfd4.setas += "2.y.x3."
    assert selfd4.setas == "2.y.x3.", "Equality test by numbered string failed."
    selfd4.setas -= "2.y"
    assert selfd4.setas == "4.x3.", "Deletion Operator failed."
    selfd4.setas = "2.y.x3."
    assert selfd4.setas["rho"] == "y", "Getitem by type failed."
    assert selfd4.setas["x"] == "T (K)", "Getitem by name failed."
    assert selfd4.setas.x == 4, "setas x attribute failed."
    assert selfd4.setas.y == [2], "setas y attribute failed."
    assert selfd4.setas.z == [], "setas z attribute failed."
    assert all(
        selfd4.setas.set == np.array([False, False, True, False, True, False, False, False])
    ), "setas.set attribute not working."
    assert all(
        selfd4.setas.not_set == np.array([True, True, False, True, False, True, True, True])
    ), "setas.set attribute not working."
    assert "x" in selfd4.setas, "set.__contains__ failed"
    del selfd4.setas["x"]
    assert "x" not in selfd4.setas, "setas del item by column type failed."
    del selfd4.setas["rho"]
    assert "y" not in selfd4.setas, "setas del by column named failed"
    selfd4.setas({"T (K)": "x", "y": "rho"})
    assert selfd4.setas == "..y.x...", "Setting setas by call with dictionary failed"


def test_failures():
    selfd = Data(path.join(path.dirname(__file__), "..", "CoreTest.dat"), setas="xy")
    with pytest.raises(AttributeError):
        selfd.setas.shape = (1, 2, 1)
    with pytest.raises(SyntaxError):
        selfd.setas(1.234)
    with pytest.raises(IndexError):
        selfd.setas({"i": 4})
    with pytest.raises(KeyError):
        selfd.setas({"x": "Wobble"})
    with pytest.raises(IndexError):
        selfd.setas({(2, 2): 1})
    with pytest.raises(IndexError):
        selfd.setas({"x": (1, 2)})
    with pytest.raises(ValueError):
        selfd.setas("ii")
    with pytest.raises(IndexError):
        selfd.setas[None]
    with pytest.raises(IndexError):
        selfd.setas[None] = None
    with pytest.raises(TypeError):
        selfd.setas + 3
    with pytest.raises(IndexError):
        selfd.setas + {"i": 2}
    with pytest.raises(TypeError):
        selfd.setas - 3.2
    with pytest.raises(IndexError):
        selfd.setas - {"i": 2}


def test_operators():
    selfd = Data(path.join(path.dirname(__file__), "..", "CoreTest.dat"), setas="xy")

    selfd.setas = "xy"
    assert not selfd.setas == "."
    assert not selfd.setas == "yx"
    assert not selfd.setas == "xyz"
    dd = selfd
    assert selfd.setas == dd.setas
    assert not selfd.setas != "xy"
    assert str(selfd.setas) == "xy"


def test_setas_dict_interface():
    selfd = Data(path.join(path.dirname(__file__), "..", "CoreTest.dat"), setas="xy")
    selfd4 = Data(path.join(__home__, "..", "sample-data", "Cu_resistivity_vs_T.txt"))

    assert list(selfd.setas.items()) == [("X-Data", "x"), ("Y-Data", "y")], "Items method of setas failed."
    assert list(selfd.setas.keys()) == ["X-Data", "Y-Data"], "setas keys method fails."

    assert list(selfd.setas.values()) == ["x", "y"], "setas values method fails."
    assert selfd.setas.get("x") == "X-Data", "Simple get with key on setas failed."
    assert selfd.setas.get("e", "Help") == "Help", "get on setas with non-existent key failed."
    assert selfd.setas.get("X-Data") == "x", "get on setas by column named failed."
    assert selfd.setas.pop("x") == "X-Data", "pop on setas failed."
    assert selfd.setas == ".y", "Residual setas after pop wrong."
    selfd.setas.clear()
    assert selfd.setas == "", "setas clear failed."
    selfd.setas.update({"x": 0, "y": 1})
    assert selfd.setas.popitem() == ("x", "X-Data"), "Popitem method failed."
    assert selfd.setas, ".y" == "residual after popitem wrong on setas."
    assert selfd.setas.setdefault("x", "X-Data") == "X-Data", "setas setdefault failed."
    assert selfd.setas == "xy", "Result after set default wrong."
    selfd4.setas = "2.y.x3."
    assert selfd4.setas["#x"] == 4, "Indexing setas by #type failed."
    assert selfd4.setas[1::2] == [".", ".", ".", "."], "Indexing setas with a slice failed."
    assert selfd4.setas[[1, 3, 5, 7]] == [".", ".", ".", "."], "Indexing setas with a slice failed."
    selfd4.setas.clear()
    selfd4.setas += {"x": "T (K)", "rho": "y"}
    assert selfd4.setas == "2.y.x3.", "Adding dictionary with type:column and column:type to setas failed."
    assert selfd4.setas.pop("x") == "T (K)" and selfd4.setas == "2.y5.", "Pop from setas failed."
    assert not selfd4.setas.pop("z", False), "Pop with non existent key and default value failed"
    selfd4.setas.update({"x": "T (K)", "rho": "y"})
    assert selfd4.setas == "2.y.x3.", "setas.update failed."


def test_setas_metadata():
    selfd = Data(path.join(path.dirname(__file__), "..", "CoreTest.dat"), setas="xy")
    selfd2 = Data(path.join(__home__, "..", "sample-data", "TDI_Format_RT.txt"))
    selfd4 = Data(path.join(__home__, "..", "sample-data", "Cu_resistivity_vs_T.txt"))

    d2 = selfd2.clone
    d2.setas += {"x": 0, "y": 1}
    assert d2.setas.to_dict() == {
        "x": "Temperature",
        "y": "Resistance",
    }, f"__iadd__ failure {repr(d2.setas.to_dict())}"
    d3 = selfd2.clone
    s = d3.setas.clone
    assert d3.setas + s == s, "Addition operator to itself is not equal to itself in setas."
    assert d3.setas - s == "...", "Subtraction operator of setas on itself not equal to empty setas."
    d3.setas = "..e"
    d2.setas.update(d3.setas)
    assert d2.setas.to_dict() == {
        "x": "Temperature",
        "y": "Resistance",
        "e": "Column 2",
    }, f"__iadd__ failure {repr(d2.setas.to_dict())}"
    auto_setas = {
        "axes": 2,
        "xcol": 0,
        "ycol": [1],
        "zcol": [],
        "ucol": [],
        "vcol": [],
        "wcol": [],
        "xerr": None,
        "yerr": [2],
        "zerr": [],
        "has_xcol": True,
        "has_xerr": False,
        "has_ycol": True,
        "has_yerr": True,
        "has_zcol": False,
        "has_zerr": False,
        "has_ucol": False,
        "has_vcol": False,
        "has_wcol": False,
        "has_axes": True,
        "has_uvw": False,
    }
    d2.setas = ""
    assert sorted(d2.setas._get_cols().items()) == sorted(auto_setas.items()), "Automatic guessing of setas failed!"
    d2.setas.clear()
    assert list(d2.setas) == ["."] * 3, "Failed to clear() setas"
    d2.setas[[0, 1, 2]] = "x", "y", "z"
    assert "".join(d2.setas) == "xyz", "Failed to set setas with a list"
    d2.setas[[0, 1, 2]] = "x"
    assert "".join(d2.setas) == "xxx", "Failed to set setas with an element from a list"
    d = selfd.clone
    d.setas = "xyz"
    assert repr(d.setas), "['x', 'y']" == "setas __repr__ failure {}".format(repr(d.setas))
    assert d.find_col(slice(5)) == [0, 1], f"findcol with a slice failed {d.find_col(slice(5))}"
    d = selfd2.clone
    d.setas = "xyz"
    assert d["Column 2", 5] == d[5, "Column 2"], "Indexing with mixed integer and string failed."
    assert selfd2.metadata.type(["User", "Timestamp"]) == [
        "String",
        "Timestamp",
    ], f"Metadata reading error {selfd2.metadata}"
    assert d.metadata.type(["User", "Timestamp"]) == ["String", "Timestamp"], "Metadata.type with slice failed"
    d.data["Column 2", :] = np.zeros(len(d))  # TODO make this work with d["Column 2",:] as well
    assert d.z.max() == 0.0 and d.z.min() == 0.0, "Failed to set Dataarray using string indexing"
    assert d.setas.x == 0 and d.setas.y == [1] and d.setas.z == [2]
    d.setas(x=1, y="Column 2")
    assert d.setas.x == 1 and d.setas.y == [2]
    selfd4.setas = "xyxy4."
    m1 = selfd4.setas._get_cols()
    m2 = selfd4.setas._get_cols(startx=2)
    assert "{xcol} {ycol}".format(**m1) == "0 [1]", f"setas._get_cols without startx failed.\n{m1}"
    assert "{xcol} {ycol}".format(**m2) == "2 [3]", f"setas._get_cols without startx failed.\n{m1}"


if __name__ == "__main__":  # Run some tests manually to allow debugging
    pytest.main(["--pdb", __file__])



# -*- coding: utf-8 -*-
"""Loader for maximus scan image."""
from pathlib import Path


from ..decorators import register_loader
from ...core.exceptions import StonerLoadError
from ..utils.maximus import read_scan, flatten_header, hdr_to_dict


@register_loader(
    patterns=[(".hdr", 16), (".xim", 16)], mime_types=("text/plain", 16), name="MaximusImage", what="Image"
)
def load_maximus_image(new_data, filename, *args, **kargs):
    """Load an ImageFile by calling the ImageArray method instead."""
    try:
        new_data.filename = filename
        pth = Path(new_data.filename)
    except TypeError as err:
        raise StonerLoadError(f"UUnable to interpret {filename} as a path like object") from err
    if pth.suffix != ".hdr":  # Passed a .xim or .xsp file in instead of the hdr file.
        pth = Path("_".join(str(pth).split("_")[:-1]) + ".hdr")
    stem = pth.parent / pth.stem

    try:
        hdr = flatten_header(hdr_to_dict(pth))
        if "Image Scan" not in hdr["ScanDefinition.Type"]:
            raise StonerLoadError("Not an Maximus Single Image File")
    except (StonerLoadError, ValueError, TypeError, IOError) as err:
        raise StonerLoadError("Error loading as Maximus File") from err
    new_data.metadata.update(hdr)
    new_data.image = read_scan(stem)[1]
    return new_data



# -*- coding: utf-8 -*-
"""Implements core image handling classes for the :mod:`Stoner.Image` package."""
__all__ = ["ImageArray", "ImageFile"]
import os
from copy import copy, deepcopy
import inspect
from importlib import import_module
from io import BytesIO as StreamIO
from warnings import warn

from PIL import Image
import numpy as np
import matplotlib.pyplot as plt
from scipy import ndimage as ndi
from skimage import (
    color,
    exposure,
    feature,
    io,
    measure,
    filters,
    graph,
    util,
    restoration,
    morphology,
    segmentation,
    transform,
)

from ..core.base import typeHintedDict, metadataObject
from ..core.exceptions import StonerLoadError, StonerUnrecognisedFormat
from ..Core import DataFile
from ..tools import isTuple, isLikeList, make_Data
from ..tools.file import file_dialog, get_file_name_type, auto_load_classes
from ..tools.decorators import class_modifier, image_file_adaptor, class_wrapper, clones
from ..compat import (
    string_types,
    get_filedialog,
    int_types,
    path_types,
)  # Some things to help with Python2 and Python3 compatibility
from .attrs import DrawProxy, MaskProxy
from .widgets import RegionSelect
from . import imagefuncs
from ..tools.classes import Options

IMAGE_FILES = [("Tiff File", "*.tif;*.tiff"), ("PNG files", "*.png", "Numpy Files", "*.npy")]

dtype_range = {
    np.bool_: (False, True),
    np.bool8: (False, True),
    np.uint8: (0, 255),
    np.uint16: (0, 65535),
    np.int8: (-128, 127),
    np.int16: (-32768, 32767),
    np.int64: (-(2**63), 2**63 - 1),
    np.uint64: (0, 2**64 - 1),
    np.int32: (-(2**31), 2**31 - 1),
    np.uint32: (0, 2**32 - 1),
    np.float16: (-1, 1),
    np.float32: (-1, 1),
    np.float64: (-1, 1),
}


def _add_core_(result, other):
    """Actually do result=result-other."""
    if isinstance(other, type(result)) and result.shape == other.shape:
        result.image += other.image
    elif isinstance(other, np.ndarray) and other.shape == result.shape:
        result.image += other
    elif isinstance(other, (int, float)):
        result.image += other
    else:
        return NotImplemented
    return result


def _floor_div_core_(result, other):
    """Actually do result=result/other."""
    # Cheat and pass through to ImageArray

    if isinstance(other, ImageFile):
        other = other.image

    result.image = result.image // other
    return result


def _div_core_(result, other):
    """Actually do result=result/other."""
    # Cheat and pass through to ImageArray

    if isinstance(other, ImageFile):
        other = other.image

    result.image = result.image / other
    return result


def _sub_core_(result, other):
    """Actually do result=result-other."""
    if isinstance(other, type(result)) and result.shape == other.shape:
        result.image -= other.image
    elif isinstance(other, np.ndarray) and other.shape == result.shape:
        result.image -= other
    elif isinstance(other, (int, float)):
        result.image -= other
    else:
        return NotImplemented
    return result


def copy_into(source: "ImageFile", dest: "ImageFile") -> "ImageFile":
    """Copy the data associated with source to dest.

    Args:
        source(ImageFile): The ImageFile object to be copied from
        dest (ImageFile): The ImageFile objrct to be changed by receiving the copiued data.

    Returns:
        The modified *dest* ImageFile.

    Unlike copying or deepcopying a ImageFile, this function preserves the class of the destination and just
    overwrites the attributes that represent the data in the ImageFile.
    """
    dest.image = source.image.clone
    for k in source._public_attrs:
        if hasattr(source, k):
            setattr(dest, k, deepcopy(getattr(source, k)))
    return dest


@class_modifier(
    [
        color,
        exposure,
        feature,
        io,
        measure,
        filters,
        filters.rank,
        graph,
        util,
        restoration,
        morphology,
        segmentation,
        transform,
    ]
)
@class_modifier([ndi], transpose=True)
@class_modifier(imagefuncs, overload=True)
class ImageArray(np.ma.MaskedArray, metadataObject):
    """A numpy array like class with a metadata parameter and pass through to skimage methods.

    ImageArray is for manipulating images stored as a 2d numpy array.
    It is built to be almost identical to a numpy array except for one extra
    parameter which is the metadata. This stores information about the image
    in a dictionary object for later retrieval.
    All standard numpy functions should work as normal and casting two types
    together should yield a ImageArray type (ie. ImageArray+np.ndarray=ImageArray)

    In addition any function from skimage should work and return a ImageArray.
    They can be called as eg. im=im.gaussian(sigma=2). Don't include the module
    name, just the function name (ie not filters.gaussian). Also omit the first
    image argument required by skimage.

    Attributes:
        metadata (:py:class:`Stoner.core.regexpDict`):
            A dictionary of metadata items associated with this image.
        filename (str):
            The name of the file from which this image was loaded.
        title (str):
            The title of the image (defaults to the filename).
        mask (:py:class:`numpy.ndarray of bool`):
            The underlying mask data of the image. Masked elements (i.e. where mask=True) are ignored for many
            image operations. Indexing them will return the mask fill value (typically NaN, ot -1 or -MAXINT)
            draw (:py:class:`Stoner.Image.attrs.DrawProxy`):
            A special object that allows the user to manipulate the image data by making use of
            :py:mod:`skimage.draw` functions as well as some additional drawing functions.
        clone (:py:class:`Stoner.ImageArry`):
            Return a duplicate copy of the current image - this allows subsequent methods to
            modify the cloned version rather than the original version.
        centre (tuple of (float,float)):
            The coordinates of the centre of the image.
        aspect (float):
            The aspect ratio (width/height) of the image.
        max_box (tuple (0,x-size,0-y-size)):
            The extent of the image size in a form suitable for use in defining a box.
        flip_h (:py:class:`ImageArray`):
            Clone the current image and then flip it horizontally (left-right).
        flip_v (:py:class:`ImageArray`):
            Clone the current image and then flip it vertically (top-bottom).
        CW (:py:class:`ImageArray`):
            Clone the current image and then rotate it 90 degrees clockwise.
        CCW (:py:class:`ImageArray`):
            Clone the current image and then rotate it 90 degrees counter-clockwise.
        T (:py:class:`ImageArray`):
            Transpose the current image
        shape (tuple (int,int)):
            Return the current shape of the image (rows, columns)
        dtype (:py:class:`numpy.dtype`):
            The current dtype of the elements of the image data.


    For clarity it should be noted that any function will not alter the current
    instance, it will clone it first then return the clone after performing the
    function on it.

    Note:
         For arrays the indexing is (row, column). However the normal way to index
         an image would be to do (horizontal, vert), which is the opposite.
         In ImageArray the coordinate system is chosen similar to skimage. y points
         down x points right and the origin is in the top left corner of the image.
         When indexing the array therefore you need to give it (y,x) coordinates
         for (row, column).::

              ----> x (column)
             |
             |
             v
             y (row)

         eg I want the 4th pixel in the horizontal direction and the 10th pixel down
         from the top I would ask for ImageArray[10,4]

         but if I want to translate the image 4 in the x direction and 10 in the y
         I would call im=im.translate((4,10))
    """

    # Proxy attributes for storing imported functions. Only do the import when needed
    _func_proxy = None

    # extra attributes for class beyond standard numpy ones

    # Default values for when we can't find the attribute already
    _defaults = {"debug": False, "_hardmask": False}

    fmts = ["png", "npy", "tiff", "tif"]

    # These will be overridden with instance attributes, but setting here allows ImageFile properties to be defined.
    debug = False
    filename = ""

    # now initialise class

    def __new__(cls, *args, **kargs):
        """Construct an ImageArray object.

        We're using __new__ rather than __init__ to imitate a numpy array as
        close as possible.
        """
        array_arg_keys = ["dtype", "copy", "order", "subok", "ndmin", "mask"]  # kwargs for array setup
        array_args = {k: kargs.pop(k) for k in array_arg_keys if k in kargs.keys()}
        user_metadata = kargs.get("metadata", {})

        if len(args) not in [0, 1]:
            raise ValueError(f"ImageArray expects 0 or 1 arguments, {len(args)} given")

        # 0 args initialisation
        if len(args) == 0:
            ret = np.empty((0, 0), dtype=float).view(cls)
            # merge the results of __new__ from emtadataObject
        else:
            # 1 args initialisation
            arg = args[0]
            loadfromfile = False
            if isinstance(arg, cls):
                ret = arg
            elif isinstance(arg, np.ndarray):
                # numpy array or ImageArray)
                if arg.ndim < 2:
                    ret = np.atleast_2d(arg).view(ImageArray)
                else:
                    ret = arg.view(ImageArray)
                kargs["metadata"] = getattr(arg, "metadata", typeHintedDict())
                kargs["metadata"].update(user_metadata)
            elif isinstance(arg, bool) and not arg:
                patterns = (("png", "*.png"), ("npy", "*.npy"))
                arg = get_filedialog(what="r", filetypes=patterns)
                if len(arg) == 0:
                    raise ValueError("No file given")
                loadfromfile = True
            elif isinstance(arg, path_types) or loadfromfile:
                # Filename- load datafile
                if not os.path.exists(arg):
                    raise ValueError(f"File path does not exist {arg}")
                ret = np.empty((0, 0), dtype=float).view(cls)
                ret = ret._load(arg, **array_args)  # pylint: disable=no-member
                kargs["metadata"] = getattr(ret, "metadata", typeHintedDict())
                kargs["metadata"].update(user_metadata)
            elif isinstance(arg, ImageFile):
                # extract the image
                ret = arg.image
                kargs["metadata"] = getattr(ret, "metadata", typeHintedDict())
                kargs["metadata"].update(user_metadata)
            else:
                try:  # try converting to a numpy array (eg a list type)
                    ret = np.asarray(arg, **array_args).view(cls)
                    if ret.dtype == "O":  # object dtype - can't deal with this
                        raise ValueError
                except ValueError as err:  # ok couldn't load from iterable, we're done
                    raise ValueError(f"No constructor for {arg}") from err

        asfloat = kargs.pop("asfloat", False) or kargs.pop(
            "convert_float", False
        )  # convert_float for back compatibility
        if asfloat and ret.dtype.kind != "f":  # convert to float type in place
            ret = ret.convert(np.float64)

        ret.__dict__["kargs"] = kargs
        return ret

    def __array_finalize__(self, obj):
        """__array_finalize__ is a  necessary functions when subclassing numpy.ndarray to fix some behaviours.

        See http://docs.scipy.org/doc/numpy-1.10.1/user/basics.subclassing.html for
        more info and examples
        Defaults below are only set when constructing an array using view
        eg np.arange(10).view(ImageArray). Otherwise filename and metadata
        attributes are just copied over (plus any other attributes set in
        _optinfo).
        """
        if not hasattr(self, "_optinfo"):
            setattr(self, "_optinfo", {"metadata": typeHintedDict({}), "filename": ""})

        kargs = self.__dict__.pop("kargs", {})  # pylint: disable=access-member-before-definition
        tmp = metadataObject.__new__(metadataObject)
        tmp.__dict__.update(self.__dict__)  # pylint: disable=access-member-before-definition
        self.__dict__ = tmp.__dict__

        # Deal with kwargs
        user_metadata = kargs.pop("metadata", {})

        _debug = kargs.pop("debug", False)
        _title = kargs.pop("title", None)

        self.metadata.update(user_metadata)

        # all constructors call array_finalise so metadata is now initialised
        self.filename = self.metadata.setdefault("Loaded from", "")
        self.debug = _debug
        self._title = _title
        self._public_attrs = {"title": str, "filename": str}
        self._mask_color = "red"
        self._mask_alpha = 0.5

        # merge the results of __new__ from emtadataObject

        if getattr(self, "debug", False):
            curframe = inspect.currentframe()
            calframe = inspect.getouterframes(curframe, 2)
            print(curframe, calframe)
        if obj is not None:
            self._optinfo.update(getattr(obj, "_optinfo", {}))

        super().__array_finalize__(obj=obj)

    def _load(self, filename, *args, **kargs):
        """Load an image from a file and return as a ImageArray."""
        cls = type(self)
        fmt = kargs.pop("fmt", os.path.splitext(filename)[1][1:])
        handlers = {"npy": cls._load_npy, "png": cls._load_png, "tiff": cls._load_tiff, "tif": cls._load_tiff}
        if fmt not in handlers:
            raise StonerLoadError(f"{fmt} is not a recognised format for loading.")
        ret = handlers[fmt](filename, **kargs)
        return ret

    @classmethod
    def _load_npy(cls, filename, **kargs):
        """Load image data from a numpy file."""
        image = np.load(filename)
        image = np.array(image, **kargs).view(cls)
        image.metadata["Loaded from"] = os.path.realpath(filename)  # pylint: disable=no-member
        image.filename = os.path.realpath(filename)
        return image

    @classmethod
    def _load_png(cls, filename, **kargs):  # pylint: disable=unused-argument
        """Create a new ImageArray from a png file."""
        with Image.open(filename, "r") as img:
            image = np.asarray(img).view(cls)
            # Since skimage.img_as_float() looks at the dtype of the array when mapping ranges, it's important to make
            # sure that we're not using too many bits to store the image in. This is a bit of a hack to reduce the
            # bit-depth...
            if np.issubdtype(image.dtype, np.integer):
                bits = np.ceil(np.log2(image.max()))
                if bits <= 8:
                    image = image.astype("uint8")
                elif bits <= 16:
                    image = image.astype("uint16")
                elif bits <= 32:
                    image = image.astype("uint32")
            for k in img.info:
                v = img.info[k]
                if v.startswith("b'"):
                    v = v.strip(" b'")
                    v = bytes(v)
                k = k.split("{")[0]
                image.metadata[k] = v
        image.metadata["Loaded from"] = os.path.realpath(filename)
        image.filename = os.path.realpath(filename)
        return image

    @classmethod
    def _load_tiff(cls, filename, **kargs):  # pylint: disable=unused-argument
        """Create a new ImageArray from a tiff file."""
        metadict = typeHintedDict({})
        with Image.open(filename, "r") as img:
            image = np.asarray(img)
            if image.ndim == 3:
                if image.shape[2] < 4:  # Need to add a dummy alpha channel
                    image = np.append(np.zeros_like(image[:, :, 0]), axis=2)
                image = image.view(dtype=np.uint32).reshape(image.shape[:-1])
            tags = img.tag_v2
            if 270 in tags:
                from json import loads

                try:
                    userdata = loads(tags[270])
                    typ = userdata.get("type", cls.__name__)
                    mod = userdata.get("module", cls.__module__)

                    mod = import_module(mod)
                    typ = getattr(mod, typ)
                    if not issubclass(typ, ImageArray):
                        raise TypeError(f"Bad type in Tiff file {typ.__name__} is not a subclass of Stoner.ImageArray")
                    metadata = userdata.get("metadata", [])
                except (ValueError, TypeError, IOError):
                    metadata = []
            else:
                metadata = []
            metadict.import_all(metadata)

            # OK now try and sort out the datatype before loading
        dtype = metadict.get(
            "ImageArray.dtype", None
        )  # if tif was previously saved by Stoner then dtype should have been added to the metadata
        # If we convert to float, it's important to make
        # sure that we're not using too many bits to store the image in.
        # This is a bit of a hack to reduce the bit-depth...
        if dtype is None:
            if np.issubdtype(image.dtype, np.integer):
                bits = np.ceil(np.log2(image.max()))
                if bits <= 8:
                    dtype = "uint8"
                elif bits <= 16:
                    dtype = "uint16"
                elif bits <= 32:
                    dtype = "uint32"
            else:
                dtype = np.dtype(image.dtype).name  # retain the loaded datatype
        try:
            image = image.astype(dtype)
        except TypeError:  # Python 2.7 can throw up a bad type error here
            pass
        image = image.view(cls)
        image.update(metadict)
        image.metadata["Loaded from"] = os.path.realpath(filename)
        image.filename = os.path.realpath(filename)
        return image

    def _box(self, *args, **kargs):
        """Construct and indexing tuple for selecting areas for cropping and boxing.

        The box can be specified as:

            - (int): a fixed number of pxiels is removed from all sides
            - (float): the central region of the image is selected
            - None: the whole image is selected
            - False: The user can select a region of interest
            - (iterable of length 4) - assumed to give 4 integers to describe a specific box
        """
        if len(args) == 0 and "box" in kargs.keys():
            args = [kargs["box"]]  # back compatibility
        elif len(args) not in (0, 1, 4):
            raise ValueError("box accepts 1 or 4 arguments, {len(args)} given.")
        if len(args) == 0 or (len(args) == 1 and args[0] is None):
            args = RegionSelect()(self)
        if len(args) == 1:
            box = args[0]
            if isinstance(box, bool) and not box:  # box=False is the same as all values
                return slice(None, None, None), slice(None, None, None)
            if isLikeList(box) and len(box) == 4:  # Full box as a list
                box = [x for x in box]
            elif isinstance(box, int):  # Take a border of n pixels out
                box = [box, self.shape[1] - box, box, self.shape[0] - box]
            elif isinstance(box, string_types):
                box = self.metadata[box]
                return self._box(*box)
            elif isinstance(box, float):  # Keep the central fraction of the image
                box = [
                    round(self.shape[1] * box / 2),
                    round(self.shape[1] * (1 - box / 2)),
                    round(self.shape[1] * box / 2),
                    round(self.shape[1] * (1 - box / 2)),
                ]
                box = list([int(x) for x in box])
            else:
                raise ValueError(f"crop accepts tuple of length 4, {len(box)} given.")
        else:
            box = list(args)
        for i, item in enumerate(box):  # replace None with max extent
            if isinstance(item, float) and 0 <= item <= 1:
                if i < 2:
                    box[i] = int(round(self.shape[1] * item))
                else:
                    box[i] = int(round(self.shape[0] * item))
            elif isinstance(item, float):
                box[i] = int(round(item))
            elif isinstance(item, int_types):
                pass
            elif item is None:
                box[i] = self.max_box[i]
            else:
                raise TypeError(f"Arguments for box should be floats, integers or None, not {type(item)}")
        return slice(box[2], box[3]), slice(box[0], box[1])

    #################################################################################################
    ################################################ Properties #####################################

    @property
    def aspect(self):
        """Return the aspect ratio (width/height) of the image."""
        return float(self.shape[1]) / self.shape[0]

    @property
    def centre(self):
        """Return the coordinates of the centre of the image."""
        return tuple(np.array(self.shape) / 2.0)

    @property
    def clone(self):
        """Duplicate the ImageFile and return the copy.

        Using .clone allows further methods to modify the clone, allowing the original immage to be unmodified.
        """
        ret = self.copy().view(type(self))
        self._optinfo["mask"] = self.mask  # Make sure we've updated our mask record
        self._optinfo["metadata"] = self.metadata  # Update metadata record
        for k, v in self._optinfo.items():
            try:
                setattr(ret, k, deepcopy(v))
            except (TypeError, ValueError, RecursionError):
                if isinstance(v, np.ndarray):
                    setattr(self, k, np.copy(v).view(v.__class__))
                else:
                    setattr(ret, k, copy(v))
        return ret

    @property
    def flat(self):
        """Return the numpy.ndarray.flat rather than a MaskedIterator."""
        return self.asarray().flat

    @property
    def max_box(self):
        """Return the maximum coordinate extent (xmin,xmax,ymin,ymax)."""
        box = (0, self.shape[1], 0, self.shape[0])
        return box

    @property
    def title(self):
        """Get a title for this image."""
        if self._title is None:
            return self.filename
        return self._title

    @title.setter
    def title(self, title):
        """Set the title of the current image."""
        if not isinstance(title, str):
            title = repr(title)
        self._title = title

    @property
    @clones
    def flip_h(self):
        """Clone the image and then mirror the image horizontally."""
        ret = self.clone[:, ::-1]
        return ret

    @property
    @clones
    def flip_v(self):
        """Clone the image and then mirror the image vertically."""
        ret = self.clone[::-1, :]
        return ret

    @property
    @clones
    def CW(self):
        """Clone the image and then rotate the imaage 90 degrees clockwise."""
        return self.clone.T[:, ::-1]

    @property
    @clones
    def CCW(self):
        """Clone the image and then rotate the imaage 90 degrees counter clockwise."""
        return self.clone.T[::-1, :]

    @property
    def draw(self):
        """Access the DrawProxy object for accessing the skimage draw sub module."""
        return DrawProxy(self, self)

    # ==============================================================================
    # OTHER SPECIAL METHODS
    # ==============================================================================

    def __getstate__(self):
        """Help with pickling ImageArrays."""
        ret = super().__getstate__()

        return {"numpy": ret, "ImageArray": {"metadata": self.metadata}}

    def __setstate__(self, state):
        """Help with pickling ImageArrays."""
        original = state.pop("numpy", tuple())
        local = state.pop("ImageArray", {})
        metadata = local.pop("metadata", {})
        super().__setstate__(original)
        self.metadata.update(metadata)

    def __delattr__(self, name):
        """Handle deleting attributes."""
        super().__delattr__(name)
        if name in self._optinfo:
            del self._optinfo[name]

    def __setattr__(self, name, value):
        """Set an attribute on the object."""
        super().__setattr__(name, value)
        # add attribute to those for copying in array_finalize. use value as
        # default.
        circ = ["_optinfo", "mask", "__dict__"]  # circular references
        proxy = ["_funcs"]  # can be reloaded for cloned arrays
        if name in circ + proxy:
            # Ignore these in clone
            pass
        else:
            self._optinfo.update({name: value})

    def __getitem__(self, index):
        """Patch indexing of strings to metadata."""
        if getattr(self, "debug", False):
            curframe = inspect.currentframe()
            calframe = inspect.getouterframes(curframe, 2)
            print(curframe, calframe)
        if isinstance(index, ImageFile) and index.image.dtype == bool:
            index = index.image
        if isinstance(index, string_types):
            return self.metadata[index]
        return super().__getitem__(index)

    def __setitem__(self, index, value):
        """Patch string index through to metadata."""
        if isinstance(index, ImageFile) and index.dtype == bool:
            index = index.image
        if isinstance(index, string_types):
            self.metadata[index] = value
        else:
            super().__setitem__(index, value)

    def __delitem__(self, index):
        """Patch indexing of strings to metadata."""
        if isinstance(index, string_types):
            del self.metadata[index]
        else:
            super().__delitem__(index)

    def save(self, filename=None, **kargs):
        """Stub method for a save function."""
        raise NotImplementedError(f"Save is not implemented in {self.__class__}")


@class_modifier(
    [
        color,
        exposure,
        feature,
        io,
        measure,
        filters,
        filters.rank,
        graph,
        util,
        restoration,
        morphology,
        segmentation,
        transform,
    ],
    adaptor=image_file_adaptor,
)
@class_modifier(
    [ndi],
    transpose=True,
    adaptor=image_file_adaptor,
)
@class_modifier(imagefuncs, overload=True, adaptor=image_file_adaptor)
@class_wrapper(target=ImageArray, exclude_below=metadataObject)
class ImageFile(metadataObject):
    """An Image file type that is analogous to :py:class:`Stoner.Data`.

    This contains metadata and an image attribute which
    is an :py:class:`Stoner.Image.ImageArray` type which subclasses numpy ndarray and
    adds lots of extra image specific processing functions.

    Attributes:
        image (:py:class:`Stoner.Image.ImageArray`):
            A :py:class:`numpy.ndarray` subclass that stores the actual image data.
        metadata (:py:class:`Stoner.core.regexpDict`):
            A dictionary of metadata items associated with this image.
        filename (str):
            The name of the file from which this image was loaded.
        title (str):
            The title of the image (defaults to the filename).
        mask (:py:class:`Stoner.Image.attrs.MaskProxy`):
            A special object that allows manipulation of the image's mask - thius allows the
            user to selectively disable regions of the image from rpocessing functions.
        draw (:py:class:`Stoner.Image.attrs.DrawProxy`):
            A special object that allows the user to manipulate the image data by making use of
            :py:mod:`skimage.draw` functions as well as some additional drawing functions.
        clone (:py:class:`Stoner.ImageFile`):
            Return a duplicate copy of the current image - this allows subsequent methods to
            modify the cloned version rather than the original version.
        centre (tuple of (int,int)):
            The coordinates of the centre of the image.
        aspect (float):
            The aspect ratio (width/height) of the image.
        max_box (tuple (0,x-size,0-y-size)):
            The extent of the image size in a form suitable for use in defining a box.
        flip_h (ImageFile):
            Clone the current image and then flip it horizontally (left-right).
        flip_v (ImageFile):
            Clone the current image and then flip it vertically (top-bottom).
        CW (ImageFile):
            Clone the current image and then rotate it 90 degrees clockwise.
        CCW (ImageFile):
            Clone the current image and then rotate it 90 degrees counter-clockwise.
        T (ImageFile):
            Transpose the current image
        shape (tuple (int,int)):
            Return the current shape of the image (rows, columns)
        dtype (:py:class:`numpy.dtype`):
            The current dtype of the elements of the image data.

    The ImageFile owned attribute is image. All other calls including metadata
    are passed through to ImageArray (so no need to inherit from metadataObject).

    Almost all calls to ImageFile are passed through to the underlying ImageArray
    logic and ImageArray can be used as a standalone class.
    However because ImageArray subclasses an ndarray it is not possible to enter
    it in place. All attributes return an array instance which needs to be reassigned.
    ImageFile owns image and so can change in place.
    The penalty is that numpy ufuncs don't return ImageFile type

    so can do::

        imfile.asfloat() #imagefile.image is updated to float type however need to do:
        imfile.image = np.abs(imfile.image)

    whereas for imarray need to do::

            imarray = imagearray.asfloat()

    but::

            np.abs(imarray) #returns ImageArray type
    """

    # pylint: disable=no-member

    _protected_attrs = ["_fromstack"]  # these won't be passed through to self.image attrs
    _patterns = ["*.png", "*.tif", "*.jpeg", "*.jpg"]
    mime_type = ["image/png", "image/jpeg", "image/tiff", "application/octet-stream"]
    priority = 32

    def __init__(self, *args, **kargs):
        """Mostly a pass through to ImageArray constructor.

        Local attribute is image. All other attributes and calls are passed
        through to image attribute.

        There is one special case of creating an ImageFile from a :py:class:`Stoner.Core.DataFile`. In this case the
        the DataFile is assumed to contain (x,y,z) data that should be converted to a map of
        z on a regular grid of x,y. The columns for the x,y,z data can be taken from the DataFile's
        :py:attr:`Stoner.Core.DataFile.setas` attribute or overridden by providing xcol, ycol and zcol keyword
        arguments. A further *shape* keyword can spewcify the shape as a tuple or "unique" to use the unique values of
        x and y or if omitted asquare grid will be interpolated.

        """
        self._image = ImageArray()  # Ensire we have the image data in place
        super().__init__(*args, **kargs)
        args = list(args)
        if len(args) == 0:
            pass
        elif len(args) > 0 and isinstance(args[0], path_types):
            try:
                copy_into(self.__class__.load(args[0], **kargs), self)
                self._public_attrs = {"title": str, "filename": str}
                self._fromstack = kargs.pop("_fromstack", False)  # for use by ImageStack
                return
            except StonerLoadError:
                args[0] = ImageArray(*args, **kargs)
        if len(args) > 0 and isinstance(args[0], ImageFile):  # Fixing type
            self._image = args[0].image
            for k in args[0]._public_attrs:
                setattr(self, k, getattr(args[0], k, None))
        elif len(args) > 0 and isinstance(args[0], np.ndarray):  # Fixing type
            self._image = ImageArray(*args, **kargs)
            if isinstance(args[0], ImageArray):
                for k in args[0]._public_attrs:
                    setattr(self, k, getattr(args[0], k, None))
        elif len(args) > 0 and isinstance(
            args[0], DataFile
        ):  # Support initing from a DataFile that defines x,y,z coordinates
            self._init_from_datafile(*args, **kargs)
        self._public_attrs = {"title": str, "filename": str}
        self._fromstack = kargs.pop("_fromstack", False)  # for use by ImageStack

    ###################################################################################################################
    ############################# Properties #### #####################################################################

    @property
    def _repr_png_(self):
        return self._repr_png_private_

    @property
    def clone(self):
        """Make a copy of this ImageFile."""
        new = type(self)(self.image.clone)
        for attr in self.__dict__:
            if callable(getattr(self, attr)) or attr in ["image", "metadata"] or attr.startswith("_"):
                continue
            try:
                setattr(new, attr, deepcopy(getattr(self, attr)))
            except NotImplementedError:  # Deepcopying failed, so just copy a reference instead
                setattr(new, attr, getattr(self, attr))
        return new

    @property
    def data(self):
        """Alias for image[:]. Equivalence to Stoner.data behaviour."""
        return self.image

    @data.setter
    def data(self, value):
        """Access the image data by data attribute."""
        self.image = value

    @property
    def draw(self):
        """Access the DrawProxy object for accessing the skimage draw sub module."""
        return DrawProxy(self.image, self)

    @property
    def image(self):
        """Access the image data."""
        return self._image

    @image.setter
    def image(self, v):
        """Ensure stored image is always an ImageArray."""
        filename = self._image.filename
        metadata = self._image.metadata
        # ensure setting image goes into the same memory block if from stack
        if (
            hasattr(self, "_fromstack")
            and self._fromstack
            and self._image.shape == v.shape
            and self._image.dtype == v.dtype
        ):
            self._image[:] = np.copy(v)
        elif isinstance(v, np.ndarray):
            self._image = np.copy(v).view(ImageArray)
        else:
            self._image = ImageArray(v)
        self.filename = filename
        self._image.metadata.update(metadata)
        self._image.metadata.update(getattr(v, "metadata", {}))

    @property
    def mask(self):
        """Get the mask of the underlying IamgeArray."""
        return MaskProxy(self)

    @mask.setter
    def mask(self, value):
        """Set the underlying ImageArray's mask."""
        if isinstance(value, ImageFile):
            value = value.image
        if isinstance(value, MaskProxy):
            value = value._mask
        self.image.mask = value

    ###################################################################################################################
    ############################# Special methods #####################################################################

    def __getitem__(self, n):
        """Pass through to ImageArray."""
        try:
            ret = self.image.__getitem__(n)
            if isinstance(ret, ImageArray) and ret.ndim == 2:
                retval = self.clone
                retval.image = ret
                return retval
            return ret
        except KeyError:
            if n not in self.metadata and n in self._image.metadata:
                self.metadata[n] = self._image.metadata[n]
            return self.metadata.__getitem__(n)

    def __setitem__(self, n, v):
        """Pass through to ImageArray."""
        if isinstance(n, string_types):
            self.metadata.__setitem__(n, v)
        else:
            self.image.__setitem__(n, v)

    def __getstate__(self):
        """Record state for pickling ImageFiles."""
        ret = copy(self.__dict__)
        ret.update({"metadata": self.metadata})
        return ret

    def __setstate__(self, state):
        """Write state for unpickling ImageFiles."""
        metadata = state.pop("metadata", {})
        self.__dict__.update(state)
        self.metadata.update(metadata)

    def __delitem__(self, n):
        """Pass through to ImageArray."""
        try:
            self.image.__delitem__(n)
        except KeyError:
            self.metadata.__delitem__(n)

    def __delattr__(self, name):
        """Handle the delete attribute code."""
        super().__delattr__(name)
        if name in self._public_attrs_real:
            del self._public_attrs_real[name]

    def __setattr__(self, n, v):
        """Handle setting attributes."""
        obj, _ = self._where_attr(n)
        if obj is None:  # This is a new attribute so note it for preserving
            obj = self
            if self._where_attr("_public_attrs_real")[0] is self:
                self._public_attrs = {n: type(v)}
        if obj is self:
            super().__setattr__(n, v)
        else:
            setattr(obj, n, v)

    def __add__(self, other):
        """Implement the subtract operator."""
        result = self.clone
        result = _add_core_(result, other)
        return result

    def __iadd__(self, other):
        """Implement the inplace subtract operator."""
        result = self
        result = _add_core_(result, other)
        return result

    def __floordiv__(self, other):
        """Implement a // operator to do XMCD calculations on a whole image."""
        if isinstance(other, ImageFile):
            if (
                hasattr(other, "polarization")
                and hasattr(self, "polarization")
                and getattr(self, "polarization") == getattr(other, "polarization")
            ):
                raise ValueError("Can only calculate and XMCD ratio from images of opposite polarization")
            if not (hasattr(other, "polarization") and hasattr(self, "polarization")) and Options().warnings:
                warn("Calculating XMCD ratio even though one or both image polarizations cannoty be determined.")
            if self.image.dtype != other.image.dtype:
                raise ValueError(
                    "Only ImageFiles with the same type of underlying image data can be used to calculate an"
                    + "XMCD ratio.Mismatch is {self.image.dtype} vs {other.image.dtype}"
                )
            if self.image.dtype.kind != "f":
                ret = self.clone.convert(float)
                other = other.clone.convert(float)
            else:
                ret = self.clone
            plus, minus = self, other
            polarization = getattr(self, "polarization", 1)
            ret.image = polarization * (plus.image - minus.image) / (plus.image + minus.image)
            return ret
        result = self
        return _floor_div_core_(result, other)

    def __truediv__(self, other):
        """Implement the divide operator."""
        result = self.clone
        result = _div_core_(result, other)
        return result

    def __itruediv__(self, other):
        """Implement the inplace divide operator."""
        result = self
        result = _div_core_(result, other)
        return result

    def __sub__(self, other):
        """Implement the subtract operator."""
        result = self.clone
        result = _sub_core_(result, other)
        return result

    def __isub__(self, other):
        """Implement the inplace subtract operator."""
        result = self
        result = _sub_core_(result, other)
        return result

    def __neg__(self):
        """Intelliegent negate function that handles unsigned integers."""
        ret = self.clone
        if self._image.dtype.kind == "u":
            for k in dtype_range:  # Have to manually look for dtype :-()
                if k == self._image.dtype:
                    break
            else:
                raise TypeError(f"Unrecognised unsigned type {self._image.dtype}, cannot negate sensibly !")
            high_val = dtype_range[k][1]
            ret.image = high_val - self.image
        else:
            ret.image = -self.image
        return ret

    def __invert__(self):
        """Equivalent to clockwise rotation."""
        return self.CW

    def __eq__(self, other):
        """Impleent and equality test."""
        if id(self) == id(other):
            ret = True  # short circuit for identity
        elif not isinstance(other, ImageFile):
            ret = False  # Shortcircuit for non equivalent types
        else:
            ret = self.metadata == other.metadata and np.all(self.image == other.image)
        return ret

    def __repr__(self):
        """Implement standard representation for text based consoles."""
        return (
            f"{self.filename}({type(self)}) of shape {self.shape} ({self.image.dtype}) and"
            + f" {len(self.metadata)} items of metadata"
        )

    ###################################################################################################################
    ############################# Private methods #####################################################################

    def _init_from_datafile(self, *args, **kargs):
        """Initialise ImageFile from DataFile defining x,y,z coordinates.

        Args:
            args[0] (DataFile):
                A :py:class:`Stoner.Core.DataFile` instance that defines x,y,z coordinates or has columns specified
                in keywords.

        Keyword Args:
            xcol (column index):
                Column in the DataFile that has the x-co-ordinate
            ycol (column index):
                Column in the data file that defines the y-cordinate
            zcol (column index):
                Column in the datafile that defines the intensity
        """
        data = make_Data(args[0])
        shape = kargs.pop("shape", "unique")

        _ = data._col_args(**kargs)
        data.setas(x=_.xcol, y=_.ycol, z=_.zcol)  # pylint: disable=not-callable
        if isinstance(shape, string_types) and shape == "unique":
            shape = (len(np.unique(data.x)), len(np.unique(data.y)))
        elif isTuple(shape, int_types, int_types):
            pass
        else:
            shape = None
        X, Y, Z = data.griddata(_.xcol, _.ycol, _.zcol, shape=shape)
        self.image = Z.view(ImageArray)
        self.metadata = deepcopy(data.metadata)
        self["x_vector"] = np.unique(X)
        self["y_vector"] = np.unique(Y)

    def _load(self, filename, *args, **kargs):
        """Load an ImageFile by calling the ImageArray method instead."""
        self._image = ImageArray(filename, *args, **kargs)
        for k in self._image._public_attrs:
            setattr(self, k, getattr(self._image, k, None))
        return self

    def _repr_png_private_(self):
        """Provide a display function for iPython/Jupyter."""
        fig = self.image.imshow(mask_color=self.mask.colour)
        plt.title(self.filename)
        data = StreamIO()
        fig.savefig(data, format="png")
        plt.close(fig)
        data.seek(0)
        ret = data.read()
        data.close()
        return ret

    def _where_attr(self, n):
        """Get the object that has the named attribute."""
        try:
            _ = super().__getattribute__(n)
            return self, _
        except AttributeError:
            try:
                _ = getattr(self._image, n)
                return self._image, _
            except AttributeError:
                return None, None

    ###################################################################################################################
    #############################  Public methods #####################################################################

    def get_filename(self, mode):
        """Force the user to choose a new filename using a system dialog box.

        Args:
            mode (string):
                The mode of file operation to be used when calling the dialog box

        Returns:
            str:
                The new filename

        Note:
            The filename attribute of the current instance is updated by this method as well.
        """
        self.filename = file_dialog(mode, self.filename, type(self), ImageFile)
        return self.filename

    @classmethod
    def load(cls, *args, **kargs):
        """Create a :py:class:`ImageFile` from file abnd guessing a better subclass if necessary.

        Args:
            filename (string or None):
                path to file to load

        Keyword Arguments:
            auto_load (bool):
                If True (default) then the load routine tries all the subclasses of :py:class:`ImageFile` in turn to
                load the file
            filetype (:py:class:`ImageFile`, str):
                If not none then tries using filetype as the loader.
            debug (bool):
                Turn on debugging when running autoload. Default *False*

        Returns:
            (ImageFile):
                A a new :py:data:`ImageFile` (or subclass thereof) instance

        Note:
            If *filetupe* is a string, then it is first tried as an exact match to a subclass name, otherwise it
            is used as a partial match and the first class in priority order is that matches is used.

            Some subclasses can be found in the :py:mod:`Stoner.formats` package.

            Each subclass is scanned in turn for a class attribute :py:attr:`Stoner.ImnageFile.priority` which governs
            the order in which they are tried. Subclasses which can make an early positive determination that a
            file has the correct format can have higher priority levels. Classes should return a suitable exception
            if they fail to load the file.

            If no class can load a file successfully then a RunttimeError exception is raised.
        """
        args = list(args)
        filename = kargs.pop("filename", args.pop(0) if len(args) > 0 else None)
        filetype = kargs.pop("filetype", None)
        auto_load = kargs.pop("auto_load", filetype is None)
        debug = kargs.pop("debug", False)

        filename, filetype = get_file_name_type(filename, filetype, DataFile)
        if auto_load:  # We're going to try every subclass we canA
            try:
                ret = auto_load_classes(filename, ImageFile, debug=debug, args=args, kargs=kargs)
            except StonerUnrecognisedFormat:
                ret = ImageFile()
                ret = ret._load(filename, *args, **kargs)
                ret["Loaded as"] = filetype.__name__
        else:
            if issubclass(filetype, ImageFile):
                ret = filetype()
                ret = ret._load(filename, *args, **kargs)
                ret["Loaded as"] = filetype.__name__
            elif filetype is None or isinstance(filetype, ImageFile):
                ret = cls()
                ret = ret._load(filename, *args, **kargs)
                ret["Loaded as"] = cls.__name__
            else:
                raise ValueError(f"Unable to load {filename}")

        for k, i in kargs.items():
            if not callable(getattr(ret, k, lambda x: False)):
                setattr(ret, k, i)
        ret._kargs = kargs
        return ret

    def save(self, filename=None, **kargs):
        """Save the image into the file 'filename'.

        Args:
            filename (string, bool or None):
                Filename to save data as, if this is None then the current filename for the object is used
                If this is not set, then then a file dialog is used. If filename is False then a file dialog is forced.

        Keyword Args:
            fmt (string or list):
                format to save data as. 'tif', 'png' or 'npy' or a list of them. If not included will guess from
                filename.

        Notes:
            Metadata will be preserved in .png and .tif format.

            fmt can be 'png', 'npy', 'tif', 'tiff'  or a list of more than one of those.
            tif is recommended since metadata is lost in .npy format but data is
            converted to integer format for png so that definition cannot be
            saved.
        """
        # catch before metadataObject tries to take over.
        self.image.save(filename, **kargs)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Test fitting mixin classes. Much of the code is also tested by the doc/samples scripts - so this is just
filling in some gaps.

Created on Sat Aug 24 22:56:59 2019

@author: phygbu
"""
import unittest
import sys
import os.path as path
import numpy as np


pth = path.dirname(__file__)
pth = path.realpath(path.join(pth, "../../"))
sys.path.insert(0, pth)
from Stoner import Data

from Stoner.analysis.fitting.mixins import _curve_fit_result


def fit(x, a, b, c):
    """Fitting function"""
    return a * x**2 + b * x + c


class AnalysisMixins_test(unittest.TestCase):

    """Path to sample Data File"""

    datadir = path.join(pth, "sample-data")

    def setUp(self):
        x_data = np.linspace(-10, 10, 101)
        y_data = 0.01 * x_data**2 + 0.3 * x_data - 2

        y_data *= np.random.normal(size=101, loc=1.0, scale=0.01)
        x_data += np.random.normal(size=101, scale=0.02)

        self.data = Data(x_data, y_data, column_headers=["X", "Y"])
        self.data.setas = "xy"

    def test_cuve_fit(self):
        for output, fmt in zip(["fit", "row", "full", "dict", "data"], [tuple, np.ndarray, tuple, dict, Data]):
            res = self.data.curve_fit(fit, p0=[0.02, 0.2, 2], output=output)
            self.assertTrue(
                isinstance(res, fmt),
                "Failed to get expected output from curve_fit for {} (got {})".format(output, type(res)),
            )

    def test_lmfit(self):
        for output, fmt in zip(["fit", "row", "full", "dict", "data"], [tuple, np.ndarray, tuple, dict, Data]):
            res = self.data.lmfit(fit, p0=[0.02, 0.2, 2], output=output)
            self.assertTrue(
                isinstance(res, fmt),
                "Failed to get expected output from lmfit for {} (got {})".format(output, type(res)),
            )

    def test_odr(self):
        for output, fmt in zip(["fit", "row", "full", "dict", "data"], [tuple, np.ndarray, tuple, dict, Data]):
            res = self.data.odr(fit, p0=[0.02, 0.2, 2], output=output)
            self.assertTrue(
                isinstance(res, fmt),
                "Failed to get expected output from idr for {} (got {})".format(output, type(res)),
            )

    def test_differential_evolution(self):
        for output, fmt in zip(["fit", "row", "full", "dict", "data"], [tuple, np.ndarray, tuple, dict, Data]):
            res = self.data.differential_evolution(fit, p0=[0.02, 0.2, 2], output=output)
            self.assertTrue(
                isinstance(res, fmt),
                "Failed to get expected output from differential_evolution for {} (got {})".format(output, type(res)),
            )


if __name__ == "__main__":  # Run some tests manually to allow debugging
    test = AnalysisMixins_test("test_cuve_fit")
    test.setUp()
    unittest.main()
    # test.test_apply()



"""Example of nDimArrhenius Fit."""

# pylint: disable=invalid-name
from numpy import linspace, ones_like
from numpy.random import normal

from Stoner import Data
from Stoner.analysis.fitting.models.tunnelling import (
    fowlerNordheim,
    FowlerNordheim,
)

# Make some data
V = linspace(-4, 4, 1000)
I = fowlerNordheim(V, 2500, 3.2, 15.0) + normal(size=len(V), scale=1e-6)
dI = ones_like(V) * 10e-6

d = Data(V, I, dI, setas="xye", column_headers=["Bias", "Current", "Noise"])

d.curve_fit(
    fowlerNordheim, p0=[2500, 3.2, 15.0], result=True, header="curve_fit"
)
d.setas = "xyey"
d.plot(fmt=["r.", "b-"])
d.annotate_fit(
    fowlerNordheim,
    x=0.2,
    y=0.6,
    prefix="fowlerNordheim",
    fontdict={"size": "x-small", "color": "blue"},
)

d.setas = "xye"
fit = FowlerNordheim()
p0 = [2500, 5.2, 15.0]
p0 = fit.guess(I, x=V)
for p, v, mi, mx in zip(
    ["A", "phi", "d"], [2500, 3.2, 15.0], [100, 1, 5], [1e4, 20.0, 30.0]
):
    p0[p].value = v
    p0[p].bounds = [mi, mx]
d.lmfit(FowlerNordheim, p0=p0, result=True, header="lmfit")
d.setas = "x...y"
d.plot(fmt="g-")
d.annotate_fit(
    fit,
    x=0.2,
    y=0.2,
    prefix="FowlerNordheim",
    fontdict={"size": "x-small", "color": "green"},
)

d.ylabel = "Current"
d.title = "Fowler-Nordheim Model test"



# -*- coding: utf-8 -*-
"""Module to work with scan files from an AttocubeSPM running Daisy."""
import re

from ...tools.file import FileManager
from ...core.exceptions import StonerLoadError
from ..decorators import register_loader

PARAM_RE = re.compile(r"^([\d\\.eE\+\-]+)\s*([\%A-Za-z]\S*)?$")
SCAN_NO = re.compile(r"SC_(\d+)")


def parabola(X, cx, cy, a, b, c):
    """Parabola in the X-Y plane for levelling an image."""
    x, y = X
    return a * (x - cx) ** 2 + b * (y - cy) ** 2 + c


def plane(X, a, b, c):
    """Plane equation for levelling an image."""
    x, y = X
    return a * x + b * y + c


@register_loader(patterns=(".txt", 32), mime_types=("text/plain", 32), name="AttocubeScanParametersFile", what="Data")
def load_attocube_parameters(new_data, filename, *args, **kargs):
    """Load the scan parameters text file as the metadata for a Data File.

    Args:
        root_name (str):
            The scan prefix e.g. SC_###

    Returns:
        new_data:
            The modififed scan stack.
    """
    new_data.filename = filename
    with FileManager(filename, "r") as parameters:
        if not parameters.readline().startswith("Daisy Parameter Snapshot"):
            raise StonerLoadError("Parameters file exists but does not have correct header")
        for line in parameters:
            if not line.strip():
                continue
            parts = [x.strip() for x in line.strip().split(":")]
            key = parts[0]
            value = ":".join(parts[1:])
            units = PARAM_RE.match(value)
            if units and units.groups()[1]:
                key += f" [{units.groups()[1]}]"
                value = units.groups()[0]
            new_data[key] = value
    return new_data



"""Re-binning data example."""

# pylint: disable=invalid-name
from Stoner import Data
from Stoner.plot.utils import errorfill

d = Data("Noisy_Data.txt", setas="xy")

d.template.fig_height = 6
d.template.fig_width = 8
d.figure(figsize=(6, 8), no_axes=True)
d.subplot(411)

e = d.bin(bins=0.05, mode="lin")
f = d.bin(bins=0.25, mode="lin")
d.setas = "xye"
g = d.bin(bins=0.05, mode="log")
h = d.bin(bins=50, mode="log")

for i, (binned, label) in enumerate(
    zip([e, f, g, h], ["0.05 Linear", "0.25 Linear", "0.05 Log", "50 log"])
):
    binned.subplot(411 + i)
    d.plot(fmt="k,", capsize=2.0)
    binned.fig = d.fig
    binned.plot(plotter=errorfill, label=label, color="red")

    d.xlim = (1, 6)
    d.ylim(-100.0, 400)
    d.title = "Bin demo" if i == 0 else ""



"""Example of PowerLaw Fit."""

# pylint: disable=invalid-name
from numpy import linspace
from numpy.random import normal

from Stoner import Data
from Stoner.analysis.fitting.models.generic import PowerLaw, powerLaw

# Make some data
T = linspace(50, 500, 101)
R = powerLaw(T, 1e-2, 0.6666666) * normal(size=len(T), scale=0.1, loc=1.0)
d = Data(T, R, setas="xy", column_headers=["T", "Rate"])

# Curve_fit on its own
d.curve_fit(powerLaw, p0=[1, 0.5], result=True, header="curve_fit")
d.setas = "xyy"
d.plot(fmt=["r.", "b-"])
d.annotate_fit(powerLaw, x=0.5, y=0.25)

# lmfit using lmfit guesses
fit = PowerLaw()
p0 = fit.guess(R, x=T)
d.lmfit(fit, p0=p0, result=True, header="lmfit")
d.setas = "x..y"
d.plot(fmt="g-")
d.annotate_fit(PowerLaw, x=0.5, y=0.05, prefix="PowerLaw")

d.title = "Powerlaw Test Fit"
d.ylabel = "Rate"
d.xlabel = "Temperature (K)"
